# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid
from numbers import Real

import numpy as np

from ..shared.utils import xy_rotation_matrix
from .grid_object import GridObject


class Octree(GridObject):
    """
    Octree mesh class that uses a tree structure such that cells
    can be subdivided it into eight octants.

    :param u_count: Number of cells along the u-axis.
    :param v_count: Number of cells along the v-axis.
    :param w_count: Number of cells along the w-axis.
    :param u_cell_size: Base cell size along the u-axis.
    :param v_cell_size: Base cell size along the v-axis.
    :param w_cell_size: Base cell size along the w-axis.
    :param octree_cells: Array defining the i, j, k position and size of each cell.
    """

    _TYPE_UID = uuid.UUID(
        fields=(0x4EA87376, 0x3ECE, 0x438B, 0xBF, 0x12, 0x3479733DED46)
    )
    __OCTREE_DTYPE = np.dtype(
        [("I", "<i4"), ("J", "<i4"), ("K", "<i4"), ("NCells", "<i4")]
    )
    _attribute_map: dict = GridObject._attribute_map.copy()
    _attribute_map.update(
        {
            "NU": "u_count",
            "NV": "v_count",
            "NW": "w_count",
            "Origin": "origin",
            "Rotation": "rotation",
            "U Cell Size": "u_cell_size",
            "V Cell Size": "v_cell_size",
            "W Cell Size": "w_cell_size",
        }
    )

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        u_count: int = 1,
        v_count: int = 1,
        w_count: int = 1,
        u_cell_size: float = 1.0,
        v_cell_size: float = 1.0,
        w_cell_size: float = 1.0,
        octree_cells: np.ndarray | list | tuple | None = None,
        **kwargs,
    ):
        self._u_count = self.validate_octree_count(u_count, "u")
        self._v_count = self.validate_octree_count(v_count, "v")
        self._w_count = self.validate_octree_count(w_count, "w")

        if octree_cells is None:
            octree_cells = self.base_refine()

        self._octree_cells: np.ndarray = self.validate_octree_cells(octree_cells)

        super().__init__(
            **kwargs,
        )

        self.u_cell_size = u_cell_size
        self.v_cell_size = v_cell_size
        self.w_cell_size = w_cell_size

    def base_refine(self) -> np.ndarray:
        """
        Refine the mesh to its base octree level resulting in a
        single cell along the shortest dimension.
        """
        # Number of octree levels allowed on each dimension
        level_u = np.log2(self.u_count)
        level_v = np.log2(self.v_count)
        level_w = np.log2(self.w_count)

        min_level = np.min([level_u, level_v, level_w])

        # Check that the refine level doesn't exceed the shortest dimension
        level = np.min([0, min_level])

        # Number of additional break to account for variable dimensions
        add_u = int(level_u - min_level)
        add_v = int(level_v - min_level)
        add_w = int(level_w - min_level)

        j, k, i = np.meshgrid(
            np.arange(0, self.v_count, 2 ** (level_v - add_v - level)),
            np.arange(0, self.w_count, 2 ** (level_w - add_w - level)),
            np.arange(0, self.u_count, 2 ** (level_u - add_u - level)),
        )

        octree_cells = np.c_[
            i.flatten(),
            j.flatten(),
            k.flatten(),
            np.ones_like(i.flatten()) * 2 ** (min_level - level),
        ]

        return octree_cells

    @property
    def centroids(self) -> np.ndarray:
        """
        Cell center locations in world coordinates of shape (n_cells, 3).

        .. code-block:: python

            centroids = [
                [x_1, y_1, z_1],
                ...,
                [x_N, y_N, z_N]
            ]
        """
        if getattr(self, "_centroids", None) is None:
            angle = np.deg2rad(self.rotation)
            rot = xy_rotation_matrix(angle)
            u_grid = (
                self.octree_cells["I"] + self.octree_cells["NCells"] / 2.0
            ) * self.u_cell_size
            v_grid = (
                self.octree_cells["J"] + self.octree_cells["NCells"] / 2.0
            ) * self.v_cell_size
            w_grid = (
                self.octree_cells["K"] + self.octree_cells["NCells"] / 2.0
            ) * self.w_cell_size
            xyz = np.c_[u_grid, v_grid, w_grid]
            self._centroids = np.dot(rot, xyz.T).T

            for ind, axis in enumerate(["x", "y", "z"]):
                self._centroids[:, ind] += self.origin[axis]

        return self._centroids

    @property
    def n_cells(self) -> int:
        """
        Total number of cells in the mesh
        """
        return self.octree_cells.shape[0]

    @property
    def octree_cells(self) -> np.ndarray:
        """
        Array defining the i, j, k position and size of each cell.
        The size defines the width of a cell in number of base cells,
        shape (:obj:`~geoh5py.objects.octree.Octree.n_cells`, 4).

         .. code-block:: python

            cells = [
                [i_1, j_1, k_1, size_1],
                ...,
                [i_N, j_N, k_N, size_N]
            ]
        """
        if self._octree_cells is None and self.on_file:
            self._octree_cells = self.workspace.fetch_array_attribute(
                self, "octree_cells"
            )

        return self._octree_cells

    @octree_cells.setter
    def octree_cells(self, octree_cells: np.ndarray | list | tuple):
        octree_cells = self.validate_octree_cells(octree_cells)
        if (
            self._octree_cells is not None
            and self._octree_cells.shape != octree_cells.shape
        ):
            raise ValueError(
                "New octree_cells array must have the same shape as the current octree_cells array."
            )
        self._octree_cells = octree_cells

        self.workspace.update_attribute(self, "octree_cells")

    @property
    def shape(self) -> tuple[np.int32, np.int32, np.int32]:
        """
        Number of cells along the u, v and w-axis.
        """
        return self.u_count, self.v_count, self.w_count

    @property
    def u_cell_size(self) -> float:
        """
        Base cell size along the u-axis.
        """
        return self._u_cell_size

    @u_cell_size.setter
    def u_cell_size(self, value: Real):
        if not isinstance(value, Real):
            raise TypeError("Attribute 'u_cell_size' must be type(float).")

        self._u_cell_size = float(value)
        self._centroids = None
        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def u_count(self) -> int:
        """
        Number of cells along u-axis.
        """
        return self._u_count

    @property
    def v_cell_size(self) -> float:
        """
        Base cell size along the v-axis.
        """
        return self._v_cell_size

    @v_cell_size.setter
    def v_cell_size(self, value: Real):
        if not isinstance(value, Real):
            raise TypeError("Attribute 'v_cell_size' must be type(float).")

        self._v_cell_size = float(value)
        self._centroids = None
        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def v_count(self) -> int:
        """
        Number of cells along v-axis.
        """
        return self._v_count

    @property
    def cell_volumes(self) -> np.ndarray:
        """
        Returns Octree cell volumes.

        Calculates nc * u_cell_size * nc * v_cell_size * nc * w_cell_size
        where nc is the number of base cells making up the octree cell as
        defined in the octree_cells array.

        :returns: Array of cell volumes in same order as octree_cells.
        """

        return (
            self.octree_cells["NCells"] ** 3
            * self.u_cell_size
            * self.v_cell_size
            * self.w_cell_size
        )

    @staticmethod
    def validate_octree_count(value: int, axis: str) -> np.int32:
        """
        Validate the number of cells along an axis are in power of 2.

        :param value: Number of cells along the axis.
        :param axis: Axis name.

        :return: Number of cells along the axis.
        """
        if not isinstance(value, (np.integer, int)):
            raise TypeError(f"Attribute '{axis}_count' must be type(int).")

        if np.log2(value) % 1.0 != 0:
            raise ValueError(
                f"Attribute '{axis}_count' must be type(int) in power of 2."
            )

        return np.int32(value)

    @classmethod
    def validate_octree_cells(
        cls, array: np.ndarray | list | tuple | None
    ) -> np.ndarray:
        """
        Validate the octree cell array.

        :param array: An array or list defining the i, j, k position
            and size of each cell.

        :return: A formatted recarray
        """
        if isinstance(array, (list, tuple)):
            array = np.array(array, ndmin=2)

        if not isinstance(array, np.ndarray):
            raise TypeError(
                "Attribute 'octree_cells' must be a list, tuple or numpy array. "
                f"Object of type {type(array)} provided."
            )

        if np.issubdtype(array.dtype, np.number):
            assert array.shape[1] == 4, (
                "'octree_cells' requires an ndarray of shape (*, 4)"
            )
            array = np.asarray(
                np.core.records.fromarrays(
                    array.T.tolist(),
                    dtype=cls.__OCTREE_DTYPE,
                )
            )
        if array.dtype != cls.__OCTREE_DTYPE:
            raise ValueError(
                f"Array of 'octree_cells' must be of dtype = {cls.__OCTREE_DTYPE}"
            )

        return array

    @property
    def w_cell_size(self) -> float:
        """
        Base cell size along the w-axis.
        """
        return self._w_cell_size

    @w_cell_size.setter
    def w_cell_size(self, value: Real):
        if not isinstance(value, Real):
            raise TypeError("Attribute 'w_cell_size' must be type(float).")

        self._w_cell_size = float(value)
        self._centroids = None
        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def w_count(self) -> int:
        """
        Number of cells along w-axis.
        """
        return self._w_count



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from pathlib import Path
from typing import Any
from warnings import warn

from .data import Data, PrimitiveTypeEnum


class FilenameData(Data):
    """
    Class for storing files as data blob.

    :param values: Name of the file.
    :param file_bytes: Binary representation of the file.
    """

    def __init__(
        self,
        values: str | None = None,
        file_bytes: bytes | None = None,
        name="GeoImageMesh_Image",
        public: bool = False,
        **kwargs,
    ):
        super().__init__(values=values, name=name, public=public, **kwargs)

        self.file_bytes = file_bytes

    @classmethod
    def primitive_type(cls) -> PrimitiveTypeEnum:
        return PrimitiveTypeEnum.FILENAME

    @property
    def file_bytes(self):
        """
        Binary blob value representation of a file.
        """
        if (
            self.values is not None
            and self.on_file
            and getattr(self, "_file_bytes", None) is None
        ):
            self._file_bytes = self.workspace.fetch_file_object(self.uid, self.values)

        return self._file_bytes

    @file_bytes.setter
    def file_bytes(self, value: bytes | None):
        if value is not None and self.values is None:
            raise AttributeError("FilenameData requires the 'values' to be set.")

        if not isinstance(value, bytes | None):
            raise TypeError(
                "Input 'file_bytes' for FilenameData must be of type 'bytes'."
            )

        self._file_bytes = value

        if self.on_file:
            self.workspace.update_attribute(self, "values")

    @property
    def file_name(self):
        """
        Binary blob value representation of a file.
        """
        warn("This method is deprecated. Use 'values' instead.", DeprecationWarning)

        return self.values

    def save_file(self, path: str | Path = Path(), name=None):
        """
        Save the file to disk.

        :param path: Directory to save the file to.
        :param name: Name given to the file.
        """
        Path(path).mkdir(exist_ok=True)
        if name is None:
            name = getattr(self, "values", "image.tiff")

        if self.file_bytes is not None:
            with open(Path(path) / name, "wb") as raw_binary:
                raw_binary.write(self.file_bytes)

    def validate_values(self, values: Any | None) -> Any:
        if not isinstance(values, str | None):
            raise ValueError("Input 'values' for FilenameData must be of type 'str'.")

        return values

    # TODO: implement specialization to access values.
    # Stored as a 1D array of 32-bit unsigned integer type (native).
    # Value map : 1D composite type array data set
    #   – Key (unsigned int)
    #   - Value (variable-length utf8 string)
    # must exist under type.
    # No data value : 0 (key is tied to value "Unknown")



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from .notype import NoTypeGroup


class RootGroup(NoTypeGroup):
    """The Root group of a workspace."""

    _default_name = "Workspace"

    def __init__(
        self,
        allow_move=False,
        allow_delete=False,
        allow_rename=False,
        **kwargs,
    ):
        super().__init__(
            allow_move=allow_move,
            allow_delete=allow_delete,
            allow_rename=allow_rename,
            parent=self,
            **kwargs,
        )

    @property
    def parent(self):
        """
        Parental entity of root is always None
        """
        return None

    @parent.setter
    def parent(self, _):
        pass



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import numpy as np

from geoh5py.objects.object_base import ObjectBase
from geoh5py.shared.utils import copy_no_reference


if TYPE_CHECKING:
    from geoh5py.groups import Group
    from geoh5py.shared.entity import EntityContainer
    from geoh5py.workspace import Workspace


class BaseSurvey(ObjectBase, ABC):
    """
    A base survey object.
    """

    __INPUT_TYPE = None
    __TYPE = None
    __UNITS = None

    @property
    @abstractmethod
    def complement(self):
        """Returns the complement object for self."""

    @property
    @abstractmethod
    def complement_reference(self):
        """Reference data linking the geometry of complement entity."""

    @complement_reference.setter
    @abstractmethod
    def complement_reference(self, complement_reference):
        """Set the reference data linking the geometry of complement entity."""

    def copy(  # pylint: disable=too-many-arguments
        self,
        parent: Group | Workspace | None = None,
        *,
        copy_children: bool = True,
        clear_cache: bool = False,
        mask: np.ndarray | None = None,
        **kwargs,
    ):
        """
        Sub-class extension of :func:`~geoh5py.objects.cell_object.CellObject.copy`.
        """
        if parent is None:
            parent = self.parent

        new_entity = self._super_copy(
            parent=parent,
            clear_cache=clear_cache,
            copy_children=copy_children,
            mask=mask,
            omit_list=self.omit_list,
            **kwargs,
        )

        if self.metadata is not None:
            new_entity.metadata = copy_no_reference(self.metadata)

        if self.complement is not None:
            self._copy_complement(
                new_entity,
                parent=parent,
                copy_children=copy_children,
                clear_cache=clear_cache,
                mask=mask,
            )

        return new_entity

    @property
    @abstractmethod
    def omit_list(self) -> tuple:
        """List of attributes to skip on copy."""

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent: EntityContainer):
        self._set_parent(parent)

        if self.complement is not None:
            self.complement._set_parent(parent)  # pylint: disable=protected-access

    @property
    @abstractmethod
    def type(self):
        """Survey element type"""

    @property
    @abstractmethod
    def type_map(self) -> dict[str, str]:
        """Map of survey element types."""

    def _copy_complement(
        self,
        new_entity: BaseSurvey,
        *,
        parent: Group | Workspace | None = None,
        copy_children: bool = True,
        clear_cache: bool = False,
        mask: np.ndarray | None = None,
    ):
        """
        Copy the complement entity to the new entity.

        :param new_entity: New entity to copy the complement to.
        :param parent: Parent group or workspace.
        :param copy_children: Copy children entities.
        :param clear_cache: Clear the cache.
        :param mask: Mask on vertices to apply to the data.
        """
        if self.complement is None:
            return None

        # Reset the mask based on Tx ID if it exists
        mask = new_entity._get_complement_mask(mask, self.complement)  # pylint: disable=protected-access
        new_complement = self.complement._super_copy(  # pylint: disable=protected-access
            parent=parent,
            omit_list=self.omit_list,
            copy_children=copy_children,
            clear_cache=clear_cache,
            mask=mask,
        )

        setattr(
            new_entity,
            self.type_map[self.complement.type],
            new_complement,
        )

        setattr(
            new_complement,
            self.type_map[new_entity.type],
            new_entity,
        )

        new_entity._renumber_reference_ids()  # pylint: disable=protected-access

        return new_complement

    @abstractmethod
    def _get_complement_mask(self, mask: np.ndarray, complement) -> np.ndarray:
        """
        Get the complement mask based on the input mask.

        :param mask: Mask on vertices or cells.
        :param complement: Complement entity.
        """

    def _renumber_reference_ids(self):
        """
        Renumber the reference IDs in the survey.
        """

    def _super_copy(
        self,
        parent: Group | Workspace | None = None,
        copy_children: bool = True,
        clear_cache: bool = False,
        mask: np.ndarray | None = None,
        **kwargs,
    ):
        """
        Call the super().copy of the class in copy_complement method.

        :return: New copy of the input entity.
        """
        new_entity = super().copy(
            parent=parent,
            copy_children=copy_children,
            clear_cache=clear_cache,
            mask=mask,
            **kwargs,
        )

        if self.complement_reference is not None:
            if not copy_children:
                self.complement_reference.copy(parent=new_entity, mask=mask)

            new_entity.complement_reference = new_entity.get_data(
                self.complement_reference.name
            )[0]

        return new_entity



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import json
from pathlib import Path
from typing import Annotated, Any
from uuid import UUID

from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    PlainSerializer,
    field_validator,
)

from geoh5py import Workspace
from geoh5py.shared.utils import fetch_active_workspace
from geoh5py.shared.validators import empty_string_to_none, none_to_empty_string
from geoh5py.ui_json.forms import BaseForm
from geoh5py.ui_json.validations import ErrorPool, UIJsonError, get_validations


OptionalPath = Annotated[
    Path | None,  # pylint: disable=unsupported-binary-operation
    BeforeValidator(empty_string_to_none),
    PlainSerializer(none_to_empty_string),
]


class BaseUIJson(BaseModel):
    """
    Base class for storing ui.json data on disk.

    :params title: Title of the application.
    :params geoh5: Path to the geoh5 file.
    :params run_command: Command to run the application.
    :params run_command_boolean: Boolean to run the command.
    :params monitoring_directory: Directory to monitor for changes.
    :params conda_environment: Conda environment to run the application.
    :params conda_environment_boolean: Boolean to run the conda environment.
    :params workspace_geoh5: Path to the workspace geoh5 file.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    version: str
    title: str
    geoh5: Path | None
    run_command: str
    monitoring_directory: OptionalPath
    conda_environment: str
    workspace_geoh5: OptionalPath

    @field_validator("geoh5", mode="after")
    @classmethod
    def workspace_path_exists(cls, path: Path):
        if not path.exists():
            raise FileNotFoundError(f"geoh5 path {path} does not exist.")
        return path

    @classmethod
    def read(cls, path: Path):
        """Create a UIJson object from file."""

        path = path.resolve()

        if not path.exists():
            raise FileNotFoundError(f"File {path} does not exist.")

        if "".join(path.suffixes[-2:]) != ".ui.json":
            raise ValueError(f"File {path} is not a .ui.json file.")

        with open(path, encoding="utf-8") as file:
            kwargs = json.load(file)

        return cls(**kwargs)

    def write(self, path: Path):
        """Write the UIJson object to file."""

        with open(path, "w", encoding="utf-8") as file:
            data = self.model_dump_json(indent=4, exclude_unset=True, by_alias=True)
            file.write(data)

    def to_params(self, workspace: Workspace | None = None) -> dict[str, Any]:
        """
        Promote, flatten and validate parameter/values dictionary.

        :param workspace: Workspace to fetch entities from.  Used for passing active
            workspaces to avoid closing and flushing data.
        """

        with fetch_active_workspace(workspace or Workspace(self.geoh5)) as geoh5:
            if geoh5 is None:
                raise ValueError("Workspace cannot be None.")

            data = {}
            errors: dict[str, Any] = {k: [] for k in self.model_fields_set}
            for field in self.model_fields_set:
                if field == "geoh5":
                    data[field] = geoh5
                    continue

                value = getattr(self, field)
                value = value.flatten() if isinstance(value, BaseForm) else value

                if isinstance(value, UUID):
                    value = self._object_or_catch(geoh5, value)

                if isinstance(value, UIJsonError):
                    errors[field].append(value)

                data[field] = value

            self.validate_data(data, errors)

        return data

    def validate_data(
        self, params: dict[str, Any] | None = None, errors: dict[str, Any] | None = None
    ):
        """
        Validate the UIJson data.

        :param params: Promoted and flattened parameters/values dictionary.  The params
            dictionary will be generated from the model values if not provided.
        :param errors: Optionally pass existing errors. Primarily for the to_params
            method.

        :raises UIJsonError: If any validations fail.
        """

        if params is None:
            self.to_params()
            return

        if errors is None:
            errors = {k: [] for k in params}

        ui_json = self.model_dump(exclude_unset=True)
        for field in self.model_fields_set:
            form = ui_json[field]
            validations = get_validations(list(form) if isinstance(form, dict) else [])
            for validation in validations:
                try:
                    validation(field, params, ui_json)
                except UIJsonError as e:
                    errors[field].append(e)

        ErrorPool(errors).throw()

    def _object_or_catch(
        self,
        workspace: Workspace,
        uuid: UUID,
    ):
        """
        Returns an object if it exists in the workspace or an error if not.

        :param workspace: Workspace to fetch entities from.
        :param uuid: UUID of the object to fetch.
        """

        obj = workspace.get_entity(uuid)
        if obj[0] is not None:
            return obj[0]

        return UIJsonError(f"Workspace does not contain an entity with uid: {uuid}.")



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid

import numpy as np

from geoh5py.data import IntegerData, ReferencedData
from geoh5py.objects.curve import Curve
from geoh5py.objects.points import Points

from .base import FEMSurvey


# pylint: disable=too-many-ancestors


class TipperSurvey(FEMSurvey):
    """
    Base tipper survey class.
    """

    __INPUT_TYPE = ["Rx and base stations"]
    _base_stations = None
    _receivers = None

    def __init__(
        self,
        base_stations: TipperBaseStations | None = None,
        **kwargs,
    ):
        self._base_stations = base_stations

        super().__init__(
            **kwargs,
        )

    @property
    def base_stations(self) -> TipperBaseStations | None:
        """The base station entity"""
        if isinstance(self, TipperBaseStations):
            return self

        if getattr(self, "_base_stations", None) is None:
            if (
                self.metadata is not None
                and "Base stations" in self.metadata["EM Dataset"]
            ):
                base_station = self.metadata["EM Dataset"]["Base stations"]
                base_station_entity = self.workspace.get_entity(base_station)[0]

                if isinstance(base_station_entity, TipperBaseStations):
                    self._base_stations = base_station_entity

        return self._base_stations

    @base_stations.setter
    def base_stations(self, base: TipperBaseStations):
        if not isinstance(base, (TipperBaseStations, type(None))):
            raise TypeError(
                f"Input `base_stations` must be of type '{TipperBaseStations}' or None"
            )

        if isinstance(self, TipperBaseStations):
            raise AttributeError(
                f"The 'base_station' attribute cannot be set on class {TipperBaseStations}."
            )

        if base.tx_id_property is not None:
            self.edit_em_metadata({"Tx ID tx property": base.tx_id_property.uid})

            if isinstance(
                self.tx_id_property, ReferencedData | IntegerData
            ) and isinstance(base.tx_id_property, ReferencedData | IntegerData):
                self.tx_id_property.entity_type = base.tx_id_property.entity_type

        self._base_stations = base
        self.edit_em_metadata({"Base stations": base.uid})

    def copy_from_extent(
        self,
        extent: np.ndarray,
        parent=None,
        *,
        copy_children: bool = True,
        clear_cache: bool = False,
        inverse: bool = False,
        **kwargs,
    ) -> TipperReceivers | TipperBaseStations | None:
        """
        Sub-class extension of :func:`~geoh5py.shared.entity.Entity.copy_from_extent`.
        """
        indices = self.mask_by_extent(extent, inverse=inverse)

        if indices is None:
            return None

        new_entity = self.copy(
            parent=parent,
            copy_children=copy_children,
            clear_cache=clear_cache,
            mask=indices,
            **kwargs,
        )

        return new_entity

    @property
    def default_input_types(self) -> list[str]:
        """Choice of survey creation types."""
        return self.__INPUT_TYPE

    @property
    def default_receiver_type(self):
        """
        :return: Transmitter class
        """
        return TipperReceivers

    @property
    def default_transmitter_type(self):
        """
        :return: Transmitter class
        """
        return type(None)

    @property
    def base_receiver_type(self):
        return Curve

    @property
    def base_transmitter_type(self):
        return Points

    @property
    def default_metadata(self) -> dict:
        """
        :return: Default unique identifier
        """
        return {
            "EM Dataset": {
                "Base stations": None,
                "Channels": [],
                "Input type": "Rx and base stations",
                "Property groups": [],
                "Receivers": None,
                "Survey type": "ZTEM",
                "Unit": "Hertz (Hz)",
            }
        }

    @property
    def default_units(self) -> list[str]:
        """Accepted time units. Must be one of "Seconds (s)",
        "Milliseconds (ms)", "Microseconds (us)" or "Nanoseconds (ns)"
        """
        return self.__UNITS

    def _format_transmitter_ids(self, values, attributes):
        """
        Format transmitter ids.

        :param values: Array of transmitter ids.
        :param attributes: Attributes dictionary for the new Data.
        """
        if self.complement is not None and self.complement.tx_id_property is not None:
            attributes["entity_type"] = self.complement.tx_id_property.entity_type
        else:
            value_map = {
                ind: f"Base station {ind}" for ind in np.unique(values.astype(np.int32))
            }
            value_map[0] = "Unknown"
            attributes.update(
                {
                    "primitive_type": "REFERENCED",
                    "value_map": value_map,
                    "association": "VERTEX",
                }
            )


class TipperReceivers(TipperSurvey, Curve):  # pylint: disable=too-many-ancestors
    """
    A z-tipper EM survey object.
    """

    _TYPE_UID = uuid.UUID("{0b639533-f35b-44d8-92a8-f70ecff3fd26}")
    __TYPE = "Receivers"
    _default_name = "Tipper rx"

    @property
    def complement(self):
        return self.base_stations

    @property
    def type(self):
        """Survey element type"""
        return self.__TYPE


class TipperBaseStations(TipperSurvey, Points):
    """
    A z-tipper EM survey object.
    """

    _TYPE_UID = uuid.UUID("{f495cd13-f09b-4a97-9212-2ea392aeb375}")
    __TYPE = "Base stations"
    _default_name = "Tipper base"
    _minimum_vertices = 1

    @property
    def complement(self):
        return self.receivers

    @property
    def type(self):
        """Survey element type"""
        return self.__TYPE



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import numpy as np

from ..shared import FLOAT_NDV
from .data import PrimitiveTypeEnum
from .numeric_data import NumericData


class FloatData(NumericData):
    """
    Data container for floats values
    """

    def format_type(self, values: np.ndarray) -> np.ndarray:
        """
        Check if the type of values is valid and coerce to type float64.
        :param values: numpy array to modify.
        :return: the formatted values.
        """
        if not np.issubdtype(values.dtype, np.number):
            raise TypeError("Values must be a numpy array of numeric values.")

        return values

    @property
    def formatted_values(self):
        values = self.ndv_values
        if values is None:
            return values

        return values

    @classmethod
    def primitive_type(cls) -> PrimitiveTypeEnum:
        return PrimitiveTypeEnum.FLOAT

    @property
    def nan_value(self):
        """
        Nan-Data-Value
        """
        return np.nan

    @property
    def ndv(self) -> float:
        """
        No-Data-Value
        """
        return FLOAT_NDV



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import json

import numpy as np
from numpy import ndarray

from ..shared.utils import as_str_if_uuid, dict_mapper
from .data import Data
from .primitive_type_enum import PrimitiveTypeEnum


def text_formating(values: None | np.ndarray | str) -> ndarray | None:
    """
    Format text values to utf-8.

    :param values: The values to format.

    :return: The formatted values.
    """
    # todo: values[0] seems dangerous here
    if values is None or isinstance(values[0], bytes):
        return values

    return np.char.encode(values, encoding="utf-8").astype("O")


class TextData(Data):
    @property
    def formatted_values(self):
        return text_formating(self.values)

    @property
    def nan_value(self):
        """
        Nan-Data-Value to be used in arrays
        """
        return ""

    @classmethod
    def primitive_type(cls) -> PrimitiveTypeEnum:
        return PrimitiveTypeEnum.TEXT

    def validate_values(
        self, values: np.ndarray | str | None
    ) -> np.ndarray | str | None:
        if isinstance(values, bytes):
            values = values.decode()

        if isinstance(values, np.ndarray) and values.dtype == object:
            values = np.array(
                [v.decode("utf-8") if isinstance(v, bytes) else v for v in values],
                dtype=str,
            )

            if self.n_values is not None and len(values) < self.n_values:
                full_array = np.full(self.n_values, self.nan_value, dtype=values.dtype)
                full_array[: len(values)] = values
                return full_array

        if (not isinstance(values, (str, type(None), np.ndarray))) or (
            isinstance(values, np.ndarray) and values.dtype.kind not in ["U", "S"]
        ):
            raise ValueError(
                f"Input 'values' for {self} must be of type {np.ndarray}  str or None."
            )

        return values


class CommentsData(Data):
    """
    Comments added to an Object or Group.
    Stored as a list of dictionaries with the following keys:

        .. code-block:: python

            comments = [
                {
                    "Author": "username",
                    "Date": "2020-05-21T10:12:15",
                    "Text": "A text comment."
                },
            ]
    """

    @property
    def formatted_values(self):
        return json.dumps(dict_mapper(self.values, [as_str_if_uuid]))

    @classmethod
    def primitive_type(cls) -> PrimitiveTypeEnum:
        return PrimitiveTypeEnum.TEXT

    def validate_values(self, values) -> dict | None:
        if isinstance(values, str):
            values = json.loads(values)

        if values is None:
            return None

        if not isinstance(values, dict):
            raise TypeError("Input 'values' for CommentsData must be a dict.")

        if "Comments" not in values:
            raise ValueError(
                "Input 'values' for CommentsData must contain key 'Comments'."
            )

        for value in values["Comments"]:
            if not isinstance(value, dict):
                raise TypeError(
                    "Error setting CommentsData with expected input of type list[dict].\n"
                    f"Input {type(values)} provided."
                )

            if not len(set(value.keys()).union({"Author", "Date", "Text"})) == 3:
                raise ValueError(
                    "Comment dictionaries must include keys 'Author', 'Date' and 'Text'.\n"
                    f"Keys {list(value.keys())} provided."
                )

        return values


class MultiTextData(Data):
    _values: np.ndarray | str | None

    @property
    def formatted_values(self):
        return text_formating(self.values)

    @property
    def nan_value(self):
        """
        Value used to represent missing data in python.
        """
        return ""

    @classmethod
    def primitive_type(cls) -> PrimitiveTypeEnum:
        return PrimitiveTypeEnum.MULTI_TEXT

    def validate_values(
        self, values: np.ndarray | str | None
    ) -> np.ndarray | str | None:
        if isinstance(values, np.ndarray) and self.n_values is not None:
            if len(values) < self.n_values:
                full_array = np.full(self.n_values, self.nan_value, dtype=values.dtype)
                full_array[: len(values)] = values
                return full_array

        elif not isinstance(values, str | type(None)):
            raise ValueError(
                f"Input 'values' for {self} must be of type {np.ndarray}  str or None."
            )

        return values



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid

from .base import Group


class MapsGroup(Group):
    """The type for the basic Container group."""

    _TYPE_UID = uuid.UUID("{4d65f8c3-a015-4c01-b411-412c0f4f0884}")
    _default_name = "Maps Group"



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from typing import Any

from geoh5py.data import Data
from geoh5py.objects import ObjectBase


class UIJsonError(Exception):
    """Exception raised for errors in the UIJson object."""

    def __init__(self, message: str):
        super().__init__(message)


class ErrorPool:  # pylint: disable=too-few-public-methods
    """Stores validation errors for all UIJson members."""

    def __init__(self, errors: dict[str, list[Exception]]):
        self.pool = errors

    def _format_error_message(self):
        """Format the error message for the UIJsonError."""

        msg = ""
        for key, errors in self.pool.items():
            if errors:
                msg += f"\t{key}:\n"
                for i, error in enumerate(errors):
                    msg += f"\t\t{i}. {error}\n"

        return msg

    def throw(self):
        """Raise the UIJsonError with detailed list of errors per parameter."""

        message = self._format_error_message()
        if message:
            message = "Collected UIJson errors:\n" + message
            raise UIJsonError(message)


def dependency_type_validation(name: str, data: dict[str, Any], params: dict[str, Any]):
    """Dependency is optional or bool type."""

    dependency = params[name]["dependency"]
    dependency_form = params[dependency]
    if "optional" not in dependency_form and not isinstance(data[dependency], bool):
        raise UIJsonError(
            f"Dependency {dependency} must be either optional or of boolean type."
        )


def mesh_type_validation(name: str, data: dict[str, Any], params: dict[str, Any]):
    """Promoted value is one of the provided mesh types."""

    mesh_types = params[name]["mesh_type"]
    obj = data[name]
    if not isinstance(obj, tuple(mesh_types)):
        raise UIJsonError(f"Object's mesh type must be one of {mesh_types}.")


def parent_validation(name: str, data: dict[str, Any], params: dict[str, Any]):
    """Data is a child of the parent object."""

    form = params[name]
    child = data[name]
    parent = data[form["parent"]]

    if not isinstance(parent, ObjectBase) or (
        isinstance(child, Data) and parent.get_entity(child.uid)[0] is None
    ):
        raise UIJsonError(f"{name} data is not a child of {form['parent']}.")



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from pathlib import Path
from uuid import UUID

from ..groups import PropertyGroup, UIJsonGroup
from ..shared import Entity
from ..workspace import Workspace


default_ui_json = {
    "version": "",
    "title": "Custom UI",
    "geoh5": None,
    "run_command": None,
    "monitoring_directory": None,
    "conda_environment": None,
    "workspace_geoh5": None,
    "out_group": None,
}
ui_validations = {
    "association": {"values": ["Vertex", "Cell", "Object"]},
    "dataGroupType": {
        "values": [
            "Multi-element",
            "3D vector",
            "Dip direction & dip",
            "Strike & dip",
        ],
    },
    "dataType": {
        "values": ["Float", "Text", "Integer", "Referenced", "Boolean", "Filename"]
    },
    "dependency": {"types": [str, type(None)]},
    "dependencyType": {"values": ["enabled", "disabled"]},
    "enabled": {"types": [bool, type(None)]},
    "group": {"types": [str, type(None)]},
    "label": {"required": True, "types": [str]},
    "main": {"types": [bool, type(None)]},
    "meshType": {"types": [str, UUID, type(None)]},
    "optional": {"types": [bool, type(None)]},
    "parent": {"types": [str, UUID, type(None)]},
    "property": {"types": [str, UUID, type(None)]},
    "value": {
        "required": True,
        "types": [str, float, int, bool, type(None), Entity, UUID, PropertyGroup],
    },
    "tooltip": {"types": [str, type(None)]},
}
base_validations = {
    "version": {"types": [str, type(None)]},
    "title": {"required": True, "types": [str]},
    "conda_environment": {
        "types": [str, type(None)],
    },
    "geoh5": {"required": True, "types": [str, Path, Workspace, type(None)]},
    "monitoring_directory": {
        "types": [str, Path, type(None)],
    },
    "run_command": {
        "types": [str, type(None)],
    },
    "workspace_geoh5": {
        "types": [str, Path, Workspace, type(None)],
    },
    "out_group": {
        "types": [str, UIJsonGroup, type(None)],
    },
}



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import numpy as np
from h5py import special_dtype


class ReferenceValueMap:
    """Maps from reference index to reference value of ReferencedData."""

    MAP_DTYPE = np.dtype([("Key", "<u4"), ("Value", special_dtype(vlen=str))])

    def __init__(
        self,
        value_map: dict[int, str] | np.ndarray | tuple,
        name: str = "Value map",
    ):
        self._map: np.ndarray = self.validate_value_map(value_map)
        self.name = name

    def __getitem__(self, item: int) -> str:
        return dict(self._map)[item]

    def __len__(self) -> int:
        return len(self._map)

    def __call__(self) -> dict:
        try:
            map_string = self._map.astype(np.dtype([("Key", "<u4"), ("Value", "U25")]))
        except UnicodeDecodeError:
            map_string = self._map

        return dict(map_string)

    @classmethod
    def validate_value_map(cls, value_map: np.ndarray | dict) -> np.ndarray:
        """
        Verify that the key and value are valid.
        It raises errors if there is an issue

        :param value_map: Array of key, value pairs.
        """
        if isinstance(value_map, tuple):
            value_map = dict(value_map)

        if isinstance(value_map, np.ndarray) and value_map.dtype.names is None:
            if value_map.ndim == 1:
                unique_set = set(value_map)
                unique_set.discard(0)
                value_map = {i + 1: str(val) for i, val in enumerate(unique_set)}

            value_map = dict(value_map)

        if isinstance(value_map, dict):
            if not np.all(np.asarray(list(value_map)) >= 0):
                raise KeyError("Key must be an positive integer")

            value_list = list(value_map.items())
            value_map = np.array(
                value_list, dtype=[("Key", "<u4"), ("Value", special_dtype(vlen=str))]
            )

            # TODO: Replace with numpy.dtypes.StringDType instead for support of variable
            # length string after moving to numpy >=2.0
            str_len = max(
                len(val) if isinstance(val, str) else len(str(val))
                for val in value_map["Value"]
            )
            value_map["Value"] = np.char.encode(
                value_map["Value"].astype(f"U{str_len}"), "utf-8"
            )

        if not isinstance(value_map, np.ndarray):
            raise TypeError("Value map must be a numpy array or dict.")

        if value_map.dtype != cls.MAP_DTYPE:
            raise ValueError(f"Array of 'value_map' must be of dtype = {cls.MAP_DTYPE}")

        return value_map

    @property
    def map(self) -> np.ndarray:
        """
        A reference array mapping values to strings.
        The keys are positive integers, and the values description.
        The key '0' is always 'Unknown'.
        """
        return self._map

    def map_values(self, values: np.ndarray) -> np.ndarray:
        """
        Map the values to the reference values.

        :param values: The values to map.

        :return: The mapped values.
        """
        mapper = np.sort(self.map, order="Key")
        indices = np.searchsorted(mapper["Key"], values)
        return mapper["Value"][indices]


BOOLEAN_VALUE_MAP = np.array(
    [(0, b"False"), (1, b"True")],
    dtype=ReferenceValueMap.MAP_DTYPE,
)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid

import numpy as np

from ..shared.utils import str2uuid
from .cell_object import CellObject


class Curve(CellObject):
    """
    Curve object defined by a series of line segments (:obj:`~geoh5py.objects.curve.Curve.cells`)
    connecting :obj:`~geoh5py.objects.object_base.ObjectBase.vertices`.

    :param current_line_id: Unique identifier of the current line.
    :param parts: Group identifiers for vertices connected by line segments as defined by the
        :obj:`~geoh5py.objects.curve.Curve.cells` property.
    :param vertices: Array of vertices as defined by :obj:`~geoh5py.objects.points.Points.vertices`.
    """

    _attribute_map: dict = CellObject._attribute_map.copy()
    _attribute_map.update(
        {
            "Current line property ID": "current_line_id",
        }
    )
    _default_name = "Curve"
    _TYPE_UID: uuid.UUID | None = uuid.UUID(
        fields=(0x6A057FDC, 0xB355, 0x11E3, 0x95, 0xBE, 0xFD84A7FFCB88)
    )
    _minimum_vertices = 2

    def __init__(  # pylint: disable="too-many-arguments"
        self,
        current_line_id: uuid.UUID | None = None,
        parts: np.ndarray | None = None,
        vertices: np.ndarray | list | tuple | None = None,
        **kwargs,
    ):
        self._parts = self.validate_parts(parts, vertices)

        super().__init__(
            vertices=vertices,
            **kwargs,
        )
        self.current_line_id = current_line_id

    @property
    def current_line_id(self) -> uuid.UUID | None:
        """
        Unique identifier of the current line.
        """
        return self._current_line_id

    @current_line_id.setter
    def current_line_id(self, value: uuid.UUID | None):
        value = str2uuid(value)

        if not isinstance(value, (uuid.UUID, type(None))):
            raise TypeError(
                f"Input current_line_id value should be of type {uuid.UUID}."
                f" {type(value)} provided"
            )

        self._current_line_id = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    def _make_cells_from_parts(self) -> np.ndarray:
        """
        Generate cells from parts.
        """
        cells = []
        for part_id in self.unique_parts:
            ind = np.where(self.parts == part_id)[0]
            cells.append(np.sort(np.c_[ind[:-1], ind[1:]], axis=0))
        return np.vstack(cells)

    @property
    def parts(self) -> np.ndarray:
        """
        Group identifiers for vertices connected by line segments as defined by the
        :obj:`~geoh5py.objects.curve.Curve.cells`
        property. The definition of the :obj:`~geoh5py.objects.curve.Curve.cells`
        property get modified by the setting of parts.
        """
        if self._parts is None:
            if len(self.cells) == 0:
                parts = np.arange(self.n_vertices, dtype=int)
            else:
                cell_parts = np.r_[
                    0, np.cumsum(self.cells[1:, 0] != self.cells[:-1, 1])
                ]
                parts = np.zeros(self.n_vertices, dtype=int)
                parts[self.cells.flatten()] = np.kron(cell_parts, np.ones(2))
            self._parts = parts

        return self._parts

    def remove_vertices(
        self, indices: list[int] | np.ndarray, clear_cache: bool = False
    ):
        """
        Safely remove vertices and cells and corresponding data entries.

        :param indices: Indices of vertices to be removed.
        :param clear_cache: Clear cache of data values.
        """
        super().remove_vertices(indices, clear_cache=clear_cache)
        self._parts = None

    @staticmethod
    def validate_parts(
        indices: list | tuple | np.ndarray | None, vertices: np.ndarray | None
    ):
        if indices is None:
            return None

        if isinstance(indices, (list | tuple)):
            indices = np.asarray(indices, dtype="int32")

        if not isinstance(indices, np.ndarray):
            raise TypeError("Parts must be a list or numpy array.")

        indices = indices.astype("int32")

        if vertices is not None and len(indices) != vertices.shape[0]:
            raise ValueError(f"Provided parts must be of shape {vertices.shape[0]}")

        return indices

    @property
    def unique_parts(self) -> list[int]:
        """
        Unique :obj:`~geoh5py.objects.curve.Curve.parts`
        identifiers.
        """
        return np.unique(self.parts).tolist()

    def validate_cells(self, indices: list | tuple | np.ndarray | None) -> np.ndarray:
        """
        Validate or generate cells made up of pairs of vertices making
            up line segments.

        :param indices: Array of indices, shape(*, 2). If None provided, the
            vertices are connected sequentially.

        :return: Array of indices defining connected vertices.
        """
        # Auto-create from parts or connect vertices sequentially
        if indices is None:
            if self._parts is not None:
                indices = self._make_cells_from_parts()
            else:
                n_segments = self.n_vertices
                indices = np.c_[
                    np.arange(0, n_segments - 1), np.arange(1, n_segments)
                ].astype("uint32")

        if isinstance(indices, (list, tuple)):
            indices = np.array(indices, ndmin=2)

        if not isinstance(indices, np.ndarray):
            raise TypeError(
                "Attribute 'cells' must be provided as type numpy.ndarray, list or tuple."
            )

        if indices.ndim != 2 or indices.shape[1] != 2:
            raise ValueError("Array of cells should be of shape (*, 2).")

        if not np.issubdtype(indices.dtype, np.integer):
            raise TypeError("Indices array must be of integer type")

        if len(indices) > 0 and np.max(indices) > (self.n_vertices - 1):
            raise ValueError("Found cell indices larger than the number of vertices.")

        return indices



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any
from uuid import UUID

from geoh5py.shared.exceptions import (
    AggregateValidationError,
    BaseValidationError,
    InCollectionValidationError,
    RequiredFormMemberValidationError,
    RequiredObjectDataValidationError,
    RequiredUIJsonParameterValidationError,
    RequiredWorkspaceObjectValidationError,
    TypeUIDValidationError,
    TypeValidationError,
    UUIDValidationError,
    ValueValidationError,
)
from geoh5py.shared.utils import SetDict, is_uuid


class Enforcer(ABC):
    """
    Base class for rule enforcers.

    :param enforcer_type: Type of enforcer.
    :param validations: Value(s) to validate parameter value against.
    """

    enforcer_type: str = ""

    def __init__(self, validations: set):
        self.validations = validations

    @abstractmethod
    def rule(self, value: Any):
        """True if 'value' adheres to enforcers rule."""

    @abstractmethod
    def enforce(self, name: str, value: Any):
        """Enforces rule on 'name' parameter's 'value'."""

    def __eq__(self, other) -> bool:
        """Equal if same type and validations."""

        is_equal = False
        if isinstance(other, type(self)):
            is_equal = other.validations == self.validations

        return is_equal

    def __str__(self):
        return f"<{type(self).__name__}> : {self.validations}"


class TypeEnforcer(Enforcer):
    """
    Enforces valid type(s).

    :param validations: Valid type(s) for parameter value.
    :raises TypeValidationError: If value is not a valid type.
    """

    enforcer_type: str = "type"

    def __init__(self, validations: set[type]):
        super().__init__(validations)

    def enforce(self, name: str, value: Any):
        """Administers rule to enforce type validation."""

        if not self.rule(value):
            raise TypeValidationError(
                name, type(value).__name__, [k.__name__ for k in self.validations]
            )

    def rule(self, value) -> bool:
        """True if value is one of the valid types."""
        return any(isinstance(value, k) for k in self.validations.union({type(None)}))


class ValueEnforcer(Enforcer):
    """
    Enforces restricted value choices.

    :param validations: Valid parameter value(s).
    :raises ValueValidationError: If value is not a valid value
        choice.
    """

    enforcer_type = "value"

    def __init__(self, validations: set[Any]):
        super().__init__(validations)

    def enforce(self, name: str, value: Any):
        """Administers rule to enforce value validation."""

        if not self.rule(value):
            raise ValueValidationError(name, value, list(self.validations))

    def rule(self, value: Any) -> bool:
        """True if value is a valid choice."""
        return value in self.validations


class TypeUIDEnforcer(Enforcer):
    """
    Enforces restricted geoh5 entity_type uid(s).

    :param validations: Valid geoh5py object type uid(s).
    :raises TypeValidationError: If value is not a valid type uid.
    """

    enforcer_type = "type_uid"

    def __init__(self, validations: set[str]):
        super().__init__(validations)

    def enforce(self, name: str, value: Any):
        """Administers rule to enforce type uid validation."""

        if not self.rule(value):
            raise TypeUIDValidationError(name, value, list(self.validations))

    def rule(self, value: Any) -> bool:
        """True if value is a valid type uid."""
        return self.validations == {""} or value.default_type_uid() in [
            UUID(k) for k in self.validations
        ]


class UUIDEnforcer(Enforcer):
    """
    Enforces valid uuid string.

    :param validations: No validations needed, can be empty set.
    :raises UUIDValidationError: If value is not a valid uuid string.
    """

    enforcer_type = "uuid"

    def __init__(self, validations=None):
        super().__init__(validations)

    def enforce(self, name: str, value: Any):
        """Administers rule to check if valid uuid."""

        if not self.rule(value):
            raise UUIDValidationError(
                name,
                value,
            )

    def rule(self, value: Any) -> bool:
        """True if value is a valid uuid string."""

        if value is None:
            return True

        return is_uuid(value)


class RequiredEnforcer(Enforcer):
    """
    Enforces required items in a collection.

    :param validations: Items that are required in the collection.
    :raises InCollectionValidationError: If collection is missing one of
        the required parameters/members.
    """

    enforcer_type = "required"
    validation_error = InCollectionValidationError

    def __init__(self, validations: set[str | tuple[str, str]]):
        super().__init__(validations)

    def enforce(self, name: str, value: Any):
        """Administers rule to check if required items in collection."""

        if not self.rule(value):
            raise self.validation_error(
                name,
                [k for k in self.validations if k not in self.collection(value)],
            )

    def rule(self, value: Any) -> bool:
        """True if all required parameters are in 'value' collection."""
        return all(k in self.collection(value) for k in self.validations)

    def collection(self, value: Any) -> list[Any]:
        """Returns the collection to check for required items."""
        return value


class RequiredUIJsonParameterEnforcer(RequiredEnforcer):
    enforcer_type = "required_uijson_parameters"
    validation_error = RequiredUIJsonParameterValidationError


class RequiredFormMemberEnforcer(RequiredEnforcer):
    enforcer_type = "required_form_members"
    validation_error = RequiredFormMemberValidationError


class RequiredWorkspaceObjectEnforcer(RequiredEnforcer):
    enforcer_type = "required_workspace_object"
    validation_error = RequiredWorkspaceObjectValidationError

    def rule(self, value: Any) -> bool:
        """True if all objects are in the workspace."""
        validations = [value[k]["value"].uid for k in self.validations]
        return all(k in self.collection(value) for k in validations)

    def collection(self, value: dict[str, Any]) -> list[UUID]:
        return list(value["geoh5"].list_entities_name)


class RequiredObjectDataEnforcer(Enforcer):
    enforcer_type = "required_object_data"
    validation_error = RequiredObjectDataValidationError

    def enforce(self, name: str, value: Any):
        """Administers rule to check if required items in collection."""

        if not self.rule(value):
            raise self.validation_error(
                name,
                [
                    k
                    for i, k in enumerate(self.validations)
                    if value[k[1]]["value"].uid not in self.collection(value)[i]
                ],
            )

    def rule(self, value: Any) -> bool:
        """True if object/data have parent/child relationship."""
        return all(
            value[k[1]]["value"].uid in self.collection(value)[i]
            for i, k in enumerate(self.validations)
        )

    def collection(self, value: dict[str, Any]) -> list[list[UUID]]:
        """Returns list of children for all parents in validations."""
        return [
            [c.uid for c in value[k[0]]["value"].children] for k in self.validations
        ]


class EnforcerPool:
    """
    Validate data on a collection of enforcers.

    :param name: Name of parameter.
    :param enforcers: List (pool) of enforcers.
    """

    enforcer_types = {
        "type": TypeEnforcer,
        "value": ValueEnforcer,
        "uuid": UUIDEnforcer,
        "type_uid": TypeUIDEnforcer,
        "required": RequiredEnforcer,
        "required_uijson_parameters": RequiredUIJsonParameterEnforcer,
        "required_form_members": RequiredFormMemberEnforcer,
        "required_workspace_object": RequiredWorkspaceObjectEnforcer,
        "required_object_data": RequiredObjectDataEnforcer,
    }

    def __init__(self, name: str, enforcers: list[Enforcer]):
        self.name = name
        self.enforcers: list[Enforcer] = enforcers
        self._errors: list[BaseValidationError] = []

    @classmethod
    def from_validations(
        cls,
        name: str,
        validations: SetDict,
    ) -> EnforcerPool:
        """
        Create enforcers pool from validations.

        :param name: Name of parameter.
        :param validations: Encodes validations as enforcer type and
            validation key value pairs.
        :param restricted_validations: 0.

        """

        return cls(name, cls._recruit(validations))

    @property
    def validations(self) -> SetDict:
        """Returns an enforcer type / validation dictionary from pool."""
        return SetDict(**{k.enforcer_type: k.validations for k in self.enforcers})

    @staticmethod
    def _recruit(validations: SetDict):
        """Recruit enforcers from validations."""
        return [EnforcerPool._recruit_enforcer(k, v) for k, v in validations.items()]

    @staticmethod
    def _recruit_enforcer(enforcer_type: str, validation: set) -> Enforcer:
        """
        Create enforcer from enforcer type and validation.

        :param enforcer_type: Type of enforcer to create.
        :param validation: Enforcer validation.
        """

        if enforcer_type not in EnforcerPool.enforcer_types:
            raise ValueError(f"Invalid enforcer type: {enforcer_type}.")

        return EnforcerPool.enforcer_types[enforcer_type](validation)

    def enforce(self, value: Any):
        """Enforce rules from all enforcers in the pool."""

        for enforcer in self.enforcers:
            self._capture_error(enforcer, value)

        self._raise_errors()

    def _capture_error(self, enforcer: Enforcer, value: Any):
        """Catch and store 'BaseValidationError's for aggregation."""
        try:
            enforcer.enforce(self.name, value)
        except BaseValidationError as err:
            self._errors.append(err)

    def _raise_errors(self):
        """Raise errors if any exist, aggregate if more than one."""
        if self._errors:
            if len(self._errors) > 1:
                raise AggregateValidationError(self.name, self._errors)
            raise self._errors.pop()



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import weakref
from typing import TypeVar
from weakref import ReferenceType


K = TypeVar("K")  # pylint: disable=invalid-name
T = TypeVar("T")  # pylint: disable=invalid-name


def remove_none_referents(some_dict: dict[K, ReferenceType]):
    """
    Removes any key from the given ``some_dict`` where the value is a reference to a deleted value
    (that is where referent of the ``weakref`` value is None).

    :param some_dict: The dictionary to be cleaned up.
    """
    dead_keys = [key for key, value in some_dict.items() if value() is None]
    for key in dead_keys:
        del some_dict[key]


def get_clean_ref(some_dict: dict[K, ReferenceType[T]], key: K) -> T | None:
    """
    Gets the referent value for the given ``key`` in a ``some_dict`` of ``weakref`` values.
    In case ``key`` points to a reference to a deleted value, remove that key from ``some_dict``
    on the fly, and returns None.

    :param some_dict: The dictionary of ``weakref`` values.
    :param key: The key

    :return: the referent value for ``key`` if found in the the dictionary, else None.
    """
    ref = some_dict.get(key, None)
    if ref is None:
        return None
    if ref() is None:
        del some_dict[key]
        return None
    return ref()


def insert_once(some_dict: dict[K, ReferenceType], key: K, value):
    """
    Check if the reference to an Entity with uuid is already in use.

    :param some_dict: Dictionary of UUID keys and weakref values.
    :param key: UUID key to be checked.
    :param value: Entity to be checked

    :return: Dictionary with clean weakref
    """
    existing_ref = some_dict.get(key, None)
    if existing_ref is not None and existing_ref() is not None:
        raise RuntimeError(f"Key '{key}' already used.")

    some_dict[key] = weakref.ref(value)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from typing import Any
from uuid import UUID

from geoh5py import Workspace
from geoh5py.groups import PropertyGroup
from geoh5py.shared.utils import SetDict
from geoh5py.ui_json.enforcers import EnforcerPool


Validation = dict[str, Any]


class Parameter:
    """
    Basic parameter to store key/value data with validation capabilities.

    :param name: Parameter name.
    :param value: The parameters value.
    :param enforcers: A collection of enforcers.
    """

    static_validations: dict[str, Any] = {}

    def __init__(self, name: str, value: Any = None):
        self.name: str = name
        self._value: Any | None = None
        self._enforcers: EnforcerPool = EnforcerPool.from_validations(
            self.name, self.validations
        )
        if value is not None:
            self.value = value

    @property
    def validations(self):
        """Returns a dictionary of static validations."""
        return SetDict(**self.static_validations)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        self._value = val
        self.validate()

    def validate(self):
        """Validates data against the pool of enforcers."""
        self._enforcers.enforce(self.value)

    def __str__(self):
        return f"<{type(self).__name__}> : '{self.name}' -> {self.value}"


class DynamicallyRestrictedParameter(Parameter):
    """Parameter whose validations are set at runtime."""

    def __init__(
        self, name: str, restrictions: Any, enforcer_type="type", value: Any = None
    ):
        self._restrictions = restrictions
        self._enforcer_type = enforcer_type
        super().__init__(name, value)

    @property
    def restrictions(self):
        if not isinstance(self._restrictions, list):
            self._restrictions = {self._restrictions}

        return self._restrictions

    @property
    def validations(self):
        return SetDict(**{self._enforcer_type: self.restrictions})


class ValueRestrictedParameter(DynamicallyRestrictedParameter):
    """Parameter with a restricted set of values."""

    def __init__(self, name: str, restrictions: Any, value: Any = None):
        super().__init__(name, restrictions, "value", value)


class TypeRestrictedParameter(DynamicallyRestrictedParameter):
    """Parameter with a restricted set of types known at runtime only."""

    def __init__(self, name: str, restrictions: list[Any], value: Any = None):
        super().__init__(name, restrictions, "type", value)


class TypeUIDRestrictedParameter(DynamicallyRestrictedParameter):
    """Parameter with a restricted set of type uids known at runtime only."""

    def __init__(self, name: str, restrictions: list[UUID], value: Any = None):
        super().__init__(name, restrictions, "type_uid", value)


class StringParameter(Parameter):
    """Parameter for string values."""

    static_validations = {"type": str}


class IntegerParameter(Parameter):
    """Parameter for integer values."""

    static_validations = {"type": int}


class FloatParameter(Parameter):
    """Parameter for float values."""

    static_validations = {"type": float}


class NumericParameter(Parameter):
    """Parameter for generic numeric values."""

    static_validations = {"type": [int, float]}


class BoolParameter(Parameter):
    """Parameter for boolean values."""

    static_validations = {"type": bool}

    def __init__(self, name: str, value: bool = False):
        super().__init__(name, value)


class StringListParameter(Parameter):
    """Parameter for list of strings."""

    static_validations = {"type": [list, str]}

    # TODO: introduce type alias handling so that TypeEnforcer(list[str], str)
    # is possible


class WorkspaceParameter(Parameter):
    """Parameter for workspace objects."""

    static_validations = {"type": Workspace}


class PropertyGroupParameter(Parameter):
    """Parameter for property group objects."""

    static_validations = {"type": PropertyGroup}



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

import logging
from typing import Any

import numpy as np
from numpy.lib import recfunctions as rfn

from ..shared.utils import min_max_scaler
from .data import Data
from .primitive_type_enum import PrimitiveTypeEnum


logger = logging.getLogger(__name__)


class Colour(Data):
    """
    Base class for the ternary colour map.

    Warning: actually, GA is storing RGBA but ignore the "A" band (full 255).
    To mimic this behavior, works with RGB only (and drop A if exists).
    """

    _allowed_parent = [
        "Points",
        "Curve",
        "BlockModel",
        "Octree",
        "Grid2D",
    ]

    _formats = [("r", "u1"), ("g", "u1"), ("b", "u1"), ("a", "u1")]

    _nan_value = np.array([(90, 90, 90, 0)], dtype=_formats)

    def __init__(self, *args, **kwargs):
        self._verify_parents(kwargs.get("parent", None))

        super().__init__(*args, **kwargs)

    @classmethod
    def _array_to_rgb(cls, values: np.ndarray) -> np.ndarray:
        """
        Convert a 2D numpy array to RGB.

        :param values: 2D numpy array with 3 or 4 bands.

        :return: A structured array with "r", "g", "b" bands
        """
        if values.ndim != 2 or values.shape[1] not in (3, 4):
            raise ValueError("Values must be a 2D numpy array containing RGB bands.")

        nan_mask = np.isnan(values).any(axis=1)

        if values.dtype != np.uint8:
            values = min_max_scaler(values, 0, 255, axis=0)

        if values.shape[1] == 3:
            values = np.column_stack(
                (values, np.full(values.shape[0], 255, dtype=np.uint8))
            )

        values = values.astype(np.uint8).view(cls._formats).reshape(-1)

        values[nan_mask] = cls._nan_value

        return values

    @classmethod
    def _structured_to_rgb(cls, values: np.ndarray) -> np.ndarray:
        """
        Convert a structured array to RGB.

        :param values: The structured array with "r", "g", "b" bands.

        :return: The RGB values.
        """

        # convert the dtype names to lower case
        values = values.astype(
            [(name.lower(), values.dtype[name]) for name in values.dtype.names]
        )

        nan_mask = np.zeros(values.shape[0], dtype=bool)

        for band in ["r", "g", "b", "a"]:
            if band not in values.dtype.names:
                if band == "a":
                    values = rfn.append_fields(
                        values,
                        names="a",
                        data=np.full(values.shape[0], 255, dtype=np.uint8),
                        usemask=False,
                    )
                else:
                    raise ValueError(
                        "Values must be a 2D numpy array containing RGB bands."
                    )

            # ensure the min and the max are between 0 and 255
            if values[band].dtype != np.uint8:
                nan_mask = np.logical_or(nan_mask, np.isnan(values[band]))
                values[band] = min_max_scaler(values[band], 0, 255).astype(np.uint8)

        values[nan_mask] = cls._nan_value

        return values[["r", "g", "b", "a"]]

    @classmethod
    def _verify_parents(cls, parent: Any):
        """
        Verify if the parent is allowed for Colour data.

        For now the objects that can use Colour data are limited.
        Raises an error if the parent is not allowed.

        :param parent: The parent object class.
        """
        if parent is not None:
            # use name to avoid circular import
            parent_classes = parent.__class__.__name__
            if parent_classes not in cls._allowed_parent:
                raise TypeError(
                    f"Parent '{parent_classes}' is not allowed for Colour data.\n"
                    f"The allowed parents are {cls._allowed_parent}"
                )

    @property
    def nan_value(self):
        return self._nan_value

    @classmethod
    def primitive_type(cls) -> PrimitiveTypeEnum:
        return PrimitiveTypeEnum.COLOUR

    def validate_values(self, values: np.ndarray | None) -> Any:
        """
        Validate values for UnknownData.

        Important Note: GA is only supporting RGB but is storing as RGBA.
        Transform values to RGB always.
        """
        if values is None:
            return values

        if not isinstance(values, np.ndarray):
            raise TypeError(f"Values must be a numpy array. Get {type(values)}")

        # convert to structure array
        if values.dtype.names is None:
            values = self._array_to_rgb(values)
        else:
            values = self._structured_to_rgb(values)

        return self.format_length(values)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid

import numpy as np

from geoh5py.data import IntegerData, ReferencedData
from geoh5py.objects.points import Points

from .base import FEMSurvey


class MTReceivers(FEMSurvey, Points):  # pylint: disable=too-many-ancestors
    """
    A magnetotellurics survey object.
    """

    _TYPE_UID = uuid.UUID("{b99bd6e5-4fe1-45a5-bd2f-75fc31f91b38}")
    __TYPE = "Receivers"
    __INPUT_TYPE = ["Rx only"]
    _default_name = "Magnetotellurics rx"

    @property
    def default_input_types(self) -> list[str]:
        """Choice of survey creation types."""
        return self.__INPUT_TYPE

    @property
    def default_metadata(self) -> dict:
        """
        :return: Default unique identifier
        """
        return {
            "EM Dataset": {
                "Channels": [],
                "Input type": "Rx only",
                "Property groups": [],
                "Receivers": None,
                "Survey type": "Magnetotellurics",
                "Unit": "Hertz (Hz)",
            }
        }

    @property
    def default_receiver_type(self):
        """
        :return: Transmitter class
        """
        return MTReceivers

    @property
    def default_transmitter_type(self):
        """
        :return: Transmitter class
        """
        return type(None)

    @property
    def base_receiver_type(self):
        return Points

    @property
    def base_transmitter_type(self):
        return type(None)

    @property
    def type(self):
        """Survey element type"""
        return self.__TYPE

    @property
    def tx_id_property(self) -> ReferencedData | IntegerData | None:
        """
        Data link between the receiver and transmitter object.
        """
        return None

    @tx_id_property.setter
    def tx_id_property(self, value: uuid.UUID | ReferencedData | np.ndarray | None):
        raise NotImplementedError(
            "Transmitter property not available for magnetotellurics"
        )



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid

import numpy as np

from ..shared.utils import xy_rotation_matrix
from .grid_object import GridObject


class BlockModel(GridObject):
    """
    Rectilinear 3D tensor mesh defined by three perpendicular axes.

    Each axis is divided into discrete intervals that define the cell dimensions.
    Nodal coordinates are determined relative to the origin and the sign of cell delimiters.
    Negative and positive cell delimiters
    are accepted to denote relative offsets from the origin.

    :param object_type: Type of object registered in the workspace.
    :param u_cell_delimiters: Nodal offsets along the u-axis relative to the origin.
    :param v_cell_delimiters: Nodal offsets along the v-axis relative to the origin.
    :param z_cell_delimiters: Nodal offsets along the z-axis relative to the origin.
    :param kwargs: Additional attributes defined by the
        :obj:`~geoh5py.objects.grid_object.GridObject` class.
    """

    _TYPE_UID = uuid.UUID(
        fields=(0xB020A277, 0x90E2, 0x4CD7, 0x84, 0xD6, 0x612EE3F25051)
    )
    _attribute_map = GridObject._attribute_map.copy()
    _attribute_map.update({"Origin": "origin", "Rotation": "rotation"})

    def __init__(
        self,
        u_cell_delimiters: np.ndarray = np.array([0.0, 1.0]),
        v_cell_delimiters: np.ndarray = np.array([0.0, 1.0]),
        z_cell_delimiters: np.ndarray = np.array([0.0, 1.0]),
        **kwargs,
    ):
        self._u_cell_delimiters = self.validate_cell_delimiters(u_cell_delimiters, "u")
        self._v_cell_delimiters = self.validate_cell_delimiters(v_cell_delimiters, "v")
        self._z_cell_delimiters = self.validate_cell_delimiters(z_cell_delimiters, "z")

        super().__init__(
            **kwargs,
        )

    def local_axis_centers(self, axis: str) -> np.ndarray:
        """
        Get the local axis centers for the block model.

        :param axis: Axis to get the centers for.
        """
        out = np.cumsum(getattr(self, f"{axis}_cell_delimiters"))
        out[2:] = out[2:] - out[:-2]
        return out[1:] / 2

    @property
    def centroids(self) -> np.ndarray:
        """
        Cell center locations in world coordinates of shape (n_cells, 3).

        .. code-block:: python

            centroids = [
                [x_1, y_1, z_1],
                ...,
                [x_N, y_N, z_N]
            ]
        """
        if getattr(self, "_centroids", None) is None:
            cell_center_u = self.local_axis_centers("u")
            cell_center_v = self.local_axis_centers("v")
            cell_center_z = self.local_axis_centers("z")
            angle = np.deg2rad(self.rotation)
            rot = xy_rotation_matrix(angle)
            u_grid, v_grid, z_grid = np.meshgrid(
                cell_center_u, cell_center_v, cell_center_z
            )
            xyz = np.c_[np.ravel(u_grid), np.ravel(v_grid), np.ravel(z_grid)]
            self._centroids = np.dot(rot, xyz.T).T

            for ind, axis in enumerate(["x", "y", "z"]):
                self._centroids[:, ind] += self.origin[axis]

        return self._centroids

    @property
    def cell_delimiters(self) -> list[np.ndarray]:
        """
        List of cell delimiters along the u, v and z-axis.
        """
        return [
            self._u_cell_delimiters,
            self._v_cell_delimiters,
            self._z_cell_delimiters,
        ]

    @property
    def shape(self) -> tuple[int, int, int]:
        """
        Number of cells along the u, v and z-axis
        """
        return self.u_cells.shape[0], self.v_cells.shape[0], self.z_cells.shape[0]

    @property
    def u_cell_delimiters(self) -> np.ndarray:
        """
        Nodal offsets along the u-axis relative to the origin.
        """
        return self._u_cell_delimiters

    @property
    def u_cells(self) -> np.ndarray:
        """
        Cell size along the u-axis.
        """
        return self.u_cell_delimiters[1:] - self.u_cell_delimiters[:-1]

    @property
    def v_cell_delimiters(self) -> np.ndarray:
        """
        Nodal offsets along the v-axis relative to the origin.
        """
        return self._v_cell_delimiters

    @property
    def v_cells(self) -> np.ndarray:
        """
        Cell size along the v-axis.
        """
        return self.v_cell_delimiters[1:] - self.v_cell_delimiters[:-1]

    @property
    def z_cell_delimiters(self) -> np.ndarray:
        """
        Nodal offsets along the u-axis relative to the origin.
        """
        return self._z_cell_delimiters

    @property
    def z_cells(self) -> np.ndarray:
        """
        Cell size along the z-axis
        """
        return self.z_cell_delimiters[1:] - self.z_cell_delimiters[:-1]

    @staticmethod
    def validate_cell_delimiters(value: np.ndarray, axis: str) -> np.ndarray:
        """
        Validate the cell delimiters for the block model.

        :param value: Cell delimiters along the axis.
        :param axis: Axis to validate the cell delimiters for.
        """
        if not isinstance(value, np.ndarray):
            raise TypeError(
                f"Attribute '{axis}_cell_delimiters' must be a numpy array."
            )

        if not np.issubdtype(value.dtype, np.number) or value.ndim != 1:
            raise ValueError(
                f"Attribute '{axis}_cell_delimiters' must be a 1D array of floats. "
                f"Provided {value}"
            )

        return value.astype(float)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import numpy as np

from .data import PrimitiveTypeEnum
from .data_type import ReferencedBooleanType
from .referenced_data import ReferencedData


class BooleanData(ReferencedData):
    """
    Data class for logical (bool) values.
    """

    @property
    def entity_type(self):
        """
        :obj:`~geoh5py.data.data_type.ReferencedBooleanType`
        """
        return self._entity_type

    @entity_type.setter
    def entity_type(self, data_type: ReferencedBooleanType):
        if not isinstance(data_type, ReferencedBooleanType):
            raise TypeError("'entity_type' must be of type ReferencedBooleanType")

        self._entity_type = data_type

        if self.on_file:
            self.workspace.update_attribute(self, "entity_type")

    def format_type(self, values: np.ndarray):
        """
        Check if the type of values is valid and coerce to type bool.
        :param values: numpy array to modify.
        :return: the formatted values.
        """
        if set(values) - {0, 1} != set():
            raise ValueError(
                f"Values provided by {self.name} are not containing only 0 or 1"
            )

        return values.astype(bool)

    @property
    def formatted_values(self):
        return super().formatted_values.astype("int8")

    @classmethod
    def primitive_type(cls) -> PrimitiveTypeEnum:
        return PrimitiveTypeEnum.BOOLEAN

    @property
    def ndv(self) -> int:
        """
        No-Data-Value
        """
        return 0



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid

import numpy as np

from .cell_object import CellObject


class Surface(CellObject):
    """
    Surface object defined by vertices and cells
    """

    _TYPE_UID = uuid.UUID(
        fields=(0xF26FEBA3, 0xADED, 0x494B, 0xB9, 0xE9, 0xB2BBCBE298E1)
    )
    _minimum_vertices = 3

    def validate_cells(self, indices: list | tuple | np.ndarray | None) -> np.ndarray:
        """
        Validate or generate cells made up of triplets of vertices making
            up triangles.

        :param indices: Array of indices, shape(*, 3). If None provided, the
            vertices are connected sequentially.

        :return: Array of indices defining connecting vertices.
        """
        if isinstance(indices, (tuple | list)):
            indices = np.array(indices, ndmin=2)

        if indices is None:
            n_vert = self.vertices.shape[0]
            indices = np.c_[
                np.arange(0, n_vert - 2), np.arange(1, n_vert - 1), np.arange(2, n_vert)
            ].astype("uint32")

        if not isinstance(indices, np.ndarray):
            raise AttributeError(
                "Attribute 'cells' must be provided as type numpy.ndarray, list or tuple."
            )

        if indices.ndim != 2 or indices.shape[-1] != 3:
            raise ValueError("Array of 'cells' should be of shape (*, 3).")

        if not np.issubdtype(indices.dtype, np.integer):
            raise TypeError("Indices array must be of integer type")

        return indices.astype(np.int32)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid
from typing import TYPE_CHECKING

from ...data import Data
from ...groups.property_group import GroupTypeEnum
from ...objects import ObjectBase
from ..utils import INV_KEY_MAP
from .concatenated import Concatenated
from .data import ConcatenatedData
from .property_group import ConcatenatedPropertyGroup


if TYPE_CHECKING:
    from ..entity import Entity
    from .concatenator import Concatenator


class ConcatenatedObject(Concatenated, ObjectBase):
    def __init__(self, **kwargs):
        if kwargs.get("parent") is None:
            raise UserWarning(
                "Creating a concatenated object must have a parent "
                "of type Concatenator."
            )

        self._parent: Concatenator
        self._property_groups: list | None = None

        super().__init__(**kwargs)

    def create_property_group(
        self,
        name=None,
        property_group_type: GroupTypeEnum | str = GroupTypeEnum.INTERVAL,
        **kwargs,
    ) -> ConcatenatedPropertyGroup:
        """
        Create a new :obj:`~geoh5py.groups.property_group.PropertyGroup`.

        :param name: Name of the property group.
        :param property_group_type: Type of property group.
        :param kwargs: Any arguments taken by the
            :obj:`~geoh5py.groups.property_group.PropertyGroup` class.

        :return: A new :obj:`~geoh5py.groups.property_group.PropertyGroup`
        """
        prop_group = ConcatenatedPropertyGroup(
            self, name=name, property_group_type=property_group_type, **kwargs
        )

        return prop_group

    def _fetch_concatenated_children(self):
        """
        Method to generate concatenated children.
        """
        attr = self.concatenator.get_concatenated_attributes(self.uid).copy()

        for key, value in attr.items():
            if "Property:" in key:
                child_data = self.workspace.get_entity(uuid.UUID(value))[0]
                if child_data is None:
                    attributes: dict = self.concatenator.get_concatenated_attributes(
                        value
                    ).copy()
                    attributes["parent"] = self
                    self.workspace.create_from_concatenation(attributes)
                elif not isinstance(child_data, ConcatenatedPropertyGroup):
                    self.add_children([child_data])

    def get_entity(self, name: str | uuid.UUID) -> list[Entity | None]:
        """
        Get a child :obj:`~geoh5py.data.data.Data` by name.

        :param name: Name of the target child data.

        :return: A list of children Data objects
        """
        if not any(child for child in self.children if isinstance(child, Data)):
            self._fetch_concatenated_children()

        if isinstance(name, uuid.UUID):
            entity_list = [child for child in self.children if child.uid == name]
        else:
            entity_list = [child for child in self.children if child.name == name]

        if not entity_list:
            return [None]

        return entity_list

    def get_data_list(self, attribute="name"):
        """
        Lazy loading of data names from concatenated attributes.
        """
        data_list = [
            attr.replace("Property:", "").replace("\u2044", "/")
            for attr in self.concatenator.get_concatenated_attributes(self.uid)
            if "Property:" in attr
        ]

        return data_list

    @property
    def parent(self) -> Concatenator:
        return self._parent

    @parent.setter
    def parent(self, parent: Concatenator):
        if not hasattr(parent, "add_children"):
            raise ValueError(
                "The 'parent' of a concatenated Object must have an "
                "'add_children' method."
            )
        parent.add_children([self])
        self._parent = parent

    @property
    def property_groups(self) -> list | None:
        if self._property_groups is None:
            property_groups = self.concatenator.fetch_values(self, "property_group_ids")

            if property_groups is None or isinstance(self, ConcatenatedData):
                property_groups = []

            for key in property_groups:
                attributes = {
                    INV_KEY_MAP.get(key, key): val
                    for key, val in self.concatenator.get_concatenated_attributes(
                        key
                    ).items()
                }
                self.fetch_property_group(**attributes, on_file=True)

            property_groups = [
                child
                for child in self.children
                if isinstance(child, ConcatenatedPropertyGroup)
            ]

            if len(property_groups) > 0:
                self._property_groups = property_groups

        return self._property_groups

    def remove_children(
        self, children: list | Concatenated | ConcatenatedPropertyGroup
    ):
        """
        Remove children from object.

        This method calls the ObjectBase parent class to remove children from the
        object children, but also deletes the children from the workspace.

        :param children: List of children to remove.
        """
        if not isinstance(children, list):
            children = [children]

        for child in children:
            if child not in self._children:
                continue

            self.concatenator.remove_entity(child)
            self._children.remove(child)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


# pylint: disable=too-many-ancestors

from __future__ import annotations

import logging
import uuid
from abc import ABC, abstractmethod
from typing import cast

import numpy as np

from ...data import Data, ReferencedData, ReferenceValueMap
from ...shared.utils import str_json_to_dict
from ..curve import Curve
from .base import BaseSurvey


logger = logging.getLogger(__name__)


class BaseElectrode(BaseSurvey, Curve, ABC):
    __TYPE = None
    __OMIT_LIST = (
        "_ab_cell_id",
        "_metadata",
        "_potential_electrodes",
        "_current_electrodes",
    )
    __TYPE_MAP = {
        "Transmitters": "current_electrodes",
        "Receivers": "potential_electrodes",
    }

    def __init__(self, **kwargs):
        self._ab_cell_id: ReferencedData | None = None
        self._current_electrodes: CurrentElectrode | None = None
        self._potential_electrodes: PotentialElectrode | None = None

        if "ab_cell_id" in kwargs:
            raise ValueError(
                "The 'ab_cell_id' must be set after instantiation of the class."
            )
        super().__init__(**kwargs)

    @property
    def ab_cell_id(self) -> ReferencedData | None:
        """
        Reference data entity mapping cells to a unique current dipole.
        """
        if getattr(self, "_ab_cell_id", None) is None:
            child = self.get_data("A-B Cell ID")
            if any(child) and isinstance(child[0], ReferencedData):
                self.ab_cell_id = child[0]

        if getattr(self, "_ab_cell_id", None) is not None:
            return self._ab_cell_id

        return None

    @ab_cell_id.setter
    def ab_cell_id(self, data: Data | np.ndarray):
        if isinstance(data, Data):
            if not isinstance(data, ReferencedData):
                raise TypeError(f"ab_cell_id must be of type {ReferencedData}")

            if data.parent.uid == self.uid:
                self._ab_cell_id = data
            else:
                self._ab_cell_id = cast(ReferencedData, data.copy(parent=self))
        else:
            if data.dtype != np.int32:
                logger.info("ab_cell_id values will be converted to type 'int32'")

            if any(self.get_data("A-B Cell ID")):
                child = self.get_data("A-B Cell ID")[0]
                if isinstance(child, ReferencedData):
                    child.values = data.astype(np.int32)
            else:
                complement = getattr(self, "complement", None)
                attributes = {
                    "values": data.astype(np.int32),
                    "association": "CELL",
                }
                if complement is not None and complement.ab_cell_id is not None:
                    attributes["entity_type"] = complement.ab_cell_id.entity_type
                else:
                    value_map = {ii: str(ii) for ii in range(data.max() + 1)}
                    value_map[0] = "Unknown"
                    attributes.update(
                        {  # type: ignore
                            "primitive_type": "REFERENCED",
                            "value_map": value_map,
                        }
                    )

                data = self.add_data({"A-B Cell ID": attributes})

                if isinstance(data, ReferencedData):
                    self._ab_cell_id = data

    @property
    def ab_map(self) -> ReferenceValueMap | None:
        """
        Get the ReferenceData.value_map of the ab_value_id
        """
        if isinstance(self.ab_cell_id, ReferencedData):
            return self.ab_cell_id.value_map
        return None

    @property
    @abstractmethod
    def complement(self) -> BaseElectrode | None:
        """
        The complement object for the current object.
        """

    @property
    def complement_reference(self):
        """Reference data linking the geometry of complement entity."""
        return self.ab_cell_id

    @complement_reference.setter
    def complement_reference(self, value: ReferencedData):
        self.ab_cell_id = value

    @property
    @abstractmethod
    def current_electrodes(self):
        """
        The associated current_electrodes (transmitters)
        """

    @property
    def omit_list(self) -> tuple:
        """
        List of attributes to omit when copying.
        """
        return self.__OMIT_LIST

    @property
    @abstractmethod
    def potential_electrodes(self):
        """
        The associated potential_electrodes (receivers)
        """

    @property
    def type_map(self) -> dict[str, str]:
        """
        Mapping of the electrode types to the associated electrode.
        """
        return self.__TYPE_MAP

    @staticmethod
    def validate_metadata(value: dict | np.ndarray | bytes) -> dict:
        if isinstance(value, np.ndarray):
            value = value[0]

        if isinstance(value, bytes):
            value = str_json_to_dict(value)

        if isinstance(value, dict):
            default_keys = ["Current Electrodes", "Potential Electrodes"]

            # check if metadata has the required keys
            if not all(key in value for key in default_keys):
                raise ValueError(f"Input metadata must have for keys {default_keys}")

        return value

    def _get_complement_mask(
        self, mask: np.ndarray, complement: BaseElectrode
    ) -> np.ndarray:
        """
        Get the complement mask based on the input mask.

        :param mask: Mask on vertices or cells.
        :param complement: Complement entity.
        """
        if self.ab_cell_id is None or complement.ab_cell_id is None:
            raise ValueError(
                "Both the object and its complement have 'ab_cell_id' set."
            )

        intersect = np.intersect1d(
            self.ab_cell_id.values,
            complement.ab_cell_id.values,
        )
        cell_mask = np.r_[[(val in intersect) for val in complement.ab_cell_id.values]]

        return cell_mask

    def _renumber_reference_ids(self):
        """
        Renumber the ab_cell_ids based on unique values of currents.
        """
        if (
            self.complement is None
            or self.complement.ab_cell_id is None
            or self.ab_cell_id is None
        ):
            raise AttributeError(
                "The 'ab_cell_id' must be set on both the object and its complement."
            )

        # Re-number the ab_cell_id
        value_map = {
            val: ind
            for ind, val in enumerate(
                np.r_[0, np.unique(self.current_electrodes.ab_cell_id.values)]
            )
        }
        new_map = {
            val: dict(self.current_electrodes.ab_cell_id.value_map.map)[val]
            for val in value_map.values()
        }
        self.complement.ab_cell_id.values = np.asarray(
            [value_map[val] for val in self.complement.ab_cell_id.values]
        )
        self.ab_cell_id.values = np.asarray(
            [value_map[val] for val in self.ab_cell_id.values]
        )
        self.ab_cell_id.entity_type.value_map = new_map  # type: ignore


class PotentialElectrode(BaseElectrode):
    """
    Ground potential electrode (receiver).
    """

    _TYPE_UID = uuid.UUID("{275ecee9-9c24-4378-bf94-65f3c5fbe163}")
    __TYPE = "Receivers"

    @property
    def complement(self) -> CurrentElectrode | None:
        return self.current_electrodes

    @property
    def current_electrodes(self):
        """
        The associated current electrode object (sources).
        """
        if getattr(self, "_current_electrodes", None) is None:
            if self.metadata is not None and "Current Electrodes" in self.metadata:
                transmitter = self.metadata["Current Electrodes"]
                transmitter_entity = self.workspace.get_entity(transmitter)[0]

                if isinstance(transmitter_entity, CurrentElectrode):
                    self._current_electrodes = transmitter_entity

        return self._current_electrodes

    @current_electrodes.setter
    def current_electrodes(self, current_electrodes: CurrentElectrode):
        if not isinstance(current_electrodes, CurrentElectrode):
            raise TypeError(
                f"Provided current_electrodes must be of type {CurrentElectrode}. "
                f"{type(current_electrodes)} provided."
            )

        metadata = {
            "Current Electrodes": current_electrodes.uid,
            "Potential Electrodes": self.uid,
        }

        self.metadata = metadata
        current_electrodes.metadata = metadata

        if isinstance(current_electrodes.ab_cell_id, ReferencedData) and isinstance(
            self.ab_cell_id, ReferencedData
        ):
            self.ab_cell_id.entity_type = current_electrodes.ab_cell_id.entity_type

    @property
    def potential_electrodes(self):
        """
        The associated potential_electrodes (receivers)
        """
        return self

    @property
    def type(self):
        """Survey element type"""
        return self.__TYPE


class CurrentElectrode(BaseElectrode):
    """
    Ground direct current electrode (transmitter).
    """

    _TYPE_UID = uuid.UUID("{9b08bb5a-300c-48fe-9007-d206f971ea92}")
    __TYPE = "Transmitters"

    def __init__(self, **kwargs):
        self._current_line_id: uuid.UUID | None = None

        super().__init__(**kwargs)

    @property
    def complement(self) -> PotentialElectrode | None:
        return self.potential_electrodes

    @property
    def current_electrodes(self):
        """
        The associated current electrode object (sources).
        """
        return self

    @current_electrodes.setter
    def current_electrodes(self, _):
        """"""

    @property
    def potential_electrodes(self) -> PotentialElectrode | None:
        """
        The associated potential_electrodes (receivers)
        """
        if getattr(self, "_potential_electrodes", None) is None:
            if self.metadata is not None and "Potential Electrodes" in self.metadata:
                potential = self.metadata["Potential Electrodes"]
                potential_entity = self.workspace.get_entity(potential)[0]

                if isinstance(potential_entity, PotentialElectrode):
                    self._potential_electrodes = potential_entity

        return self._potential_electrodes

    @potential_electrodes.setter
    def potential_electrodes(self, potential_electrodes: PotentialElectrode):
        if not isinstance(potential_electrodes, PotentialElectrode):
            raise TypeError(
                f"Provided potential_electrodes must be of type {PotentialElectrode}. "
                f"{type(potential_electrodes)} provided."
            )

        metadata = {
            "Current Electrodes": self.uid,
            "Potential Electrodes": potential_electrodes.uid,
        }

        self.metadata = metadata
        potential_electrodes.metadata = metadata

        if isinstance(potential_electrodes.ab_cell_id, ReferencedData) and isinstance(
            self.ab_cell_id, ReferencedData
        ):
            potential_electrodes.ab_cell_id.entity_type = self.ab_cell_id.entity_type

    def add_default_ab_cell_id(self):
        """
        Utility function to set ab_cell_id's based on curve cells.
        """
        if getattr(self, "cells", None) is None or self.n_cells is None:
            raise AttributeError(
                "Cells must be set before assigning default ab_cell_id"
            )

        data = np.arange(self.n_cells) + 1
        value_map = {ii: str(ii) for ii in range(self.n_cells + 1)}
        value_map[0] = "Unknown"
        ab_cell_id = self.add_data(
            {
                "A-B Cell ID": {
                    "values": data,
                    "association": "CELL",
                    "primitive_type": "REFERENCED",
                    "value_map": value_map,
                }
            }
        )

        if not isinstance(ab_cell_id, ReferencedData):
            raise UserWarning("Could not create 'A-B Cell ID' data.")

        ab_cell_id.entity_type.name = "A-B"
        self._ab_cell_id = ab_cell_id

    @property
    def type(self):
        """Survey element type"""
        return self.__TYPE



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid
from json import loads
from typing import TYPE_CHECKING

import numpy as np
from pydantic import (
    AliasChoices,
    AliasGenerator,
    BaseModel,
    ConfigDict,
    Field,
    field_validator,
)
from pydantic.alias_generators import to_pascal, to_snake

from .points import Points


if TYPE_CHECKING:
    from numpy import ndarray


class TextEntry(BaseModel):
    """
    Core parameters for text mesh data.

    :param starting_point: Starting point of the text.
    :param annotation: Annotation ID.
    :param color: Color of the text stored as a hex string.
    :param direction: Normal vector defining the direction of the text.
    :param font: Font type.
    :param font_size: Font size.
    :param layer_name: Layer name.
    :param normal: Normal vector of the text.
    :param obliquing_angle: Obliquing angle of the text.
    :param rotation_angle: Rotation angle of the text.
    :param text: Text string displayed.
    :param width_scale_factor: Width scale factor of the text.
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(  # type: ignore
            validation_alias=to_snake,
            serialization_alias=to_pascal,
        ),
        validate_assignment=True,
    )

    starting_point: str
    annotation: int = 0
    color: str = "#ffffff"
    direction: str = "{1,0,0}"
    font: str = "builtin_simple"
    font_size: int = 10
    layer_name: str = ""
    normal: str = "{0,0,1}"
    obliquing_angle: int = 0
    rotation_angle: int = 0
    text: str = "<no-data>"
    width_scale_factor: float = 1.0

    @field_validator("starting_point", "normal", "direction", mode="before")
    @classmethod
    def array_to_str(cls, float_list: np.ndarray | list) -> str:
        """
        Convert a list of floats to a string representation.
        """
        if isinstance(float_list, tuple) and len(float_list) == 1:
            float_list = float_list[0]

        if isinstance(float_list, np.ndarray | list):
            float_list = "{" + ",".join([str(val) for val in float_list]) + "}"

        return float_list


class TextMesh(BaseModel):
    """
    Text mesh data.

    :param data: List of text entries with parameters.
    """

    data: list[TextEntry] = Field(
        validation_alias=AliasChoices("data", "Text Data"),
        serialization_alias="Text Data",
    )


class TextObject(Points):
    """
    Text object for annotated text labels in viewport.

    :param vertices: Vertices of the text object.
    :param text_mesh_data: Text mesh data as a JSON string or dictionary.
    """

    _attribute_map: dict = Points._attribute_map.copy()
    _attribute_map.update({"TextMesh Data": "text_mesh_data"})
    _default_name = "Text Object"
    _TYPE_UID: uuid.UUID = uuid.UUID("{c00905d1-bc3b-4d12-9f93-07fcf1450270}")

    def __init__(
        self,
        vertices: np.ndarray | list | tuple | None = None,
        text_mesh_data: str | None = None,
        **kwargs,
    ):
        self._text_mesh_data: TextMesh

        super().__init__(vertices=vertices, text_mesh_data=text_mesh_data, **kwargs)

    def copy(
        self,
        parent=None,
        *,
        copy_children: bool = True,
        clear_cache: bool = False,
        mask: np.ndarray | None = None,
        **kwargs,
    ) -> Points:
        """
        Sub-class extension of :func:`~geoh5py.objects.points.Points.copy`.
        """
        mask = self.validate_mask(mask)
        if mask is not None:
            text_data = TextMesh(
                data=[
                    elem
                    for elem, logic in zip(self.text_mesh_data.data, mask, strict=False)
                    if logic
                ]
            )
            kwargs.update({"text_mesh_data": text_data})

        new_entity = super().copy(
            parent=parent,
            copy_children=copy_children,
            clear_cache=clear_cache,
            mask=mask,
            **kwargs,
        )
        return new_entity

    @property
    def text_mesh_data(self) -> TextMesh:
        """
        Text mesh data.
        """

        return self._text_mesh_data

    @text_mesh_data.setter
    def text_mesh_data(self, value: str | dict | TextMesh | None):
        """
        Set the text mesh data.
        """
        if value is None:
            self._text_mesh_data = TextMesh(
                data=[
                    TextEntry(text="<no-data>", starting_point=vert)
                    for vert in self.vertices
                ]
            )
            return

        if isinstance(value, str):
            t_m_data = loads(value)
        else:
            t_m_data = value

        if isinstance(t_m_data, dict):
            t_m_data = TextMesh(**t_m_data)

        if not isinstance(t_m_data, TextMesh):
            raise TypeError("The 'Text Data' must be a dictionary or a JSON string.")

        if len(t_m_data.data) != self.n_vertices:
            raise ValueError(
                "The 'Text Data' dictionary must contain a list of len('n_vertices')."
            )

        self._text_mesh_data = t_m_data

    def _validate_value_length(self, values) -> list:
        """
        Validate the length of the values.

        :param values: Values to validate stored as a list or numpy array.
        """
        if not isinstance(values, (list, np.ndarray)):
            values = [values] * self.n_vertices

        if len(values) != self.n_vertices:
            raise ValueError(
                "The 'Text Data' entries must contain a list of len('n_vertices')."
            )
        return values

    def __setattr__(self, key, value):
        """
        Set the attribute value.
        """
        if key in TextEntry.model_fields:
            value = self._validate_value_length(value)

            for entry, elem in zip(self.text_mesh_data.data, value, strict=False):
                setattr(entry, key, elem)

            if self.on_file:
                self.workspace.update_attribute(self, "attributes")

        else:
            super().__setattr__(key, value)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import json
import uuid
from typing import Any

import h5py
import numpy as np

from ..shared import FLOAT_NDV, fetch_h5_handle
from ..shared.utils import (
    INV_KEY_MAP,
    KEY_MAP,
    as_str_if_utf8_bytes,
    as_str_if_uuid,
    str2uuid,
    str_json_to_dict,
)


class H5Reader:
    """
    Class to read information from a geoh5 file.
    """

    @classmethod
    def fetch_attributes(
        cls,
        file: str | h5py.File,
        uid: uuid.UUID,
        entity_type: str,
    ) -> tuple[dict, dict, dict] | tuple[None, None, None]:
        """
        Get attributes of an :obj:`~geoh5py.shared.entity.Entity`.

        :param file: :obj:`h5py.File` or name of the target geoh5 file
        :param uid: Unique identifier
        :param entity_type: Type of entity from
            'group', 'data', 'object', 'group_type', 'data_type', 'object_type'

        Returns
        -------
        attributes: :obj:`dict` of attributes for the :obj:`~geoh5py.shared.entity.Entity`
        type_attributes: :obj:`dict` of attributes for the :obj:`~geoh5py.shared.entity.EntityType`
        property_groups: :obj:`dict` of data :obj:`uuid.UUID`
        """
        with fetch_h5_handle(file) as h5file:
            name = list(h5file)[0]

            entity_type = cls.format_type_string(entity_type)

            if entity_type == "Root":
                entity = h5file[name].get(entity_type)
            else:
                entity = h5file[name][entity_type].get(as_str_if_uuid(uid))

            if entity is None:
                return None, None, None

            attributes: dict = {}
            type_attributes: dict = {}
            property_groups: dict = {}

            for key, value in entity.attrs.items():
                attributes[INV_KEY_MAP.get(key, key)] = value

            # TODO Use lazy pointer to data
            if entity_type != "Data":
                for key, value in entity.items():
                    if key.lower() in ["metadata", "options"]:
                        attributes[INV_KEY_MAP[key]] = cls.fetch_metadata(
                            file, uid, entity_type, key
                        )
                    elif (
                        key in INV_KEY_MAP
                        and isinstance(value, h5py.Dataset)
                        and value.ndim > 0
                    ):
                        attributes[INV_KEY_MAP[key]] = value[:]

            if "Type" in entity:
                type_attributes = cls.fetch_type_attributes(entity["Type"])
            # Check if the entity has property_group
            if "PropertyGroups" in entity:
                property_groups = cls.fetch_property_groups(file, uid)

        return attributes, type_attributes, property_groups

    @classmethod
    def fetch_array_attribute(
        cls, file: str | h5py.File, uid: uuid.UUID, entity_type: str, key: str
    ) -> np.ndarray | None:
        """
        Get an entity attribute stores as array such as
        :obj:`~geoh5py.objects.object_base.ObjectBase.cells`.

        :param file: :obj:`h5py.File` or name of the target geoh5 file
        :param uid: Unique identifier of the target object.
        :param entity_type: Group type to fetch entity from.
        :param key: Field attribute name.

        :return cells: :obj:`numpy.ndarray` of :obj:`int`.
        """
        with fetch_h5_handle(file) as h5file:
            name = list(h5file)[0]
            label = KEY_MAP.get(key, key)

            if "types" in entity_type:
                root_handle = h5file[name]["Types"][entity_type]
            else:
                root_handle = h5file[name][entity_type]

            try:
                values = root_handle[as_str_if_uuid(uid)][label][:]
                return values
            except KeyError:
                return None

    @classmethod
    def fetch_children(
        cls, file: str | h5py.File, uid: uuid.UUID, entity_type: str
    ) -> dict:
        """
        Get :obj:`~geoh5py.shared.entity.Entity.children` of an
        :obj:`~geoh5py.shared.entity.Entity`.

        :param file: :obj:`h5py.File` or name of the target geoh5 file
        :param uid: Unique identifier
        :param entity_type: Type of entity from
            'group', 'data', 'object', 'group_type', 'data_type', 'object_type'


        :return children: [{uuid: type}, ... ]
            List of dictionaries for the children uid and type
        """
        with fetch_h5_handle(file) as h5file:
            name = list(h5file)[0]
            children: dict = {}
            entity_type = cls.format_type_string(entity_type)

            if (
                entity_type not in h5file[name]
                or as_str_if_uuid(uid) not in h5file[name][entity_type]
            ):
                return children

            entity = h5file[name][entity_type][as_str_if_uuid(uid)]

            for child_type, child_list in entity.items():
                if child_type in ["Type", "PropertyGroups", "Concatenated Data"]:
                    continue

                if isinstance(child_list, h5py.Group):
                    for uid_str in child_list:
                        children[str2uuid(uid_str)] = child_type.replace(
                            "s", ""
                        ).lower()

        return children

    @classmethod
    def fetch_concatenated_values(
        cls,
        file: str | h5py.File,
        uid: uuid.UUID,
        entity_type: str,
        label: str,
    ) -> tuple | None:
        """
        Get :obj:`~geoh5py.shared.entity.Entity.children` values of concatenated group.

        :param file: :obj:`h5py.File` or name of the target geoh5 file
        :param uid: Unique identifier
        :param entity_type: Type of entity from
            'group', 'data', 'object', 'group_type', 'data_type', 'object_type'
        :param label: Group identifier for the attribute requested.

        :return children: [{uuid: type}, ... ]
            List of dictionaries for the children uid and type
        """
        with fetch_h5_handle(file) as h5file:
            name = list(h5file)[0]
            entity_type = cls.format_type_string(entity_type)
            label = KEY_MAP.get(label, label)

            try:
                group = h5file[name][entity_type][as_str_if_uuid(uid)][
                    "Concatenated Data"
                ]
                indices = group["Index"].get(label.replace("/", "\u2044"))
                if indices is None:
                    return None

                array = group["Data"].get(label.replace("/", "\u2044"))
                if array is None:
                    array = group.get(label.replace("/", "\u2044"))

                attribute = array[:]

                if attribute.dtype in [float, "float64", "float32"]:
                    attribute[attribute == FLOAT_NDV] = np.nan

                if (
                    attribute.dtype == object
                    and len(attribute) > 0
                    and isinstance(attribute[0], bytes)
                ):
                    attribute = np.char.decode(attribute.astype(np.bytes_), "UTF-8")

                return attribute, indices[:]

            except KeyError:
                return None

    @classmethod
    def fetch_concatenated_attributes(
        cls,
        file: str | h5py.File,
        uid: uuid.UUID,
        entity_type: str,
        label: str,
    ) -> list | dict | None:
        """
        Get 'Attributes', 'Data' or 'Index' from Concatenator group.

        :param file: :obj:`h5py.File` or name of the target geoh5 file
        :param uid: Unique identifier
        :param entity_type: Type of entity from
            'group', 'data', 'object', 'group_type', 'data_type', 'object_type'
        :param label: Group identifier for the attribute requested.

        :return children: [{uuid: type}, ... ]
            List of dictionaries for the children uid and type
        """
        with fetch_h5_handle(file) as h5file:
            name = list(h5file)[0]
            entity_type = cls.format_type_string(entity_type)
            label = KEY_MAP.get(label, label)

            try:
                group = h5file[name][entity_type][as_str_if_uuid(uid)][
                    "Concatenated Data"
                ]
                if label == "Attributes":
                    attribute = group[label][()]
                    if isinstance(attribute, np.ndarray):
                        attribute = attribute[0]

                    return json.loads(as_str_if_utf8_bytes(attribute))

                if label == "Attributes Jsons":
                    attribute = group[label][()]
                    return {
                        "Attributes": [
                            json.loads(as_str_if_utf8_bytes(val)) for val in attribute
                        ]
                    }

                return list(group[label])

            except KeyError:
                return None

    @classmethod
    def fetch_metadata(
        cls,
        file: str | h5py.File,
        uid: uuid.UUID,
        entity_type: str = "Objects",
        argument: str = "Metadata",
    ) -> str | dict | None:
        """
        Fetch text of dictionary type attributes of an entity.

        :param file: Target h5 file.
        :param uid: Unique identifier of the target Entity.
        :param entity_type: Base type of the target Entity.
        :param argument: Label name of the dictionary.
        """
        with fetch_h5_handle(file) as h5file:
            name = list(h5file)[0]

            try:
                value = np.r_[h5file[name][entity_type][as_str_if_uuid(uid)][argument]][
                    0
                ]

            except KeyError:
                return None

        return str_json_to_dict(value)

    @classmethod
    def fetch_project_attributes(cls, file: str | h5py.File) -> dict[Any, Any]:
        """
        Get attributes of an :obj:`~geoh5py.shared.entity.Entity`.

        :param file: :obj:`h5py.File` or name of the target geoh5 file

        :return attributes: :obj:`dict` of attributes.
        """
        with fetch_h5_handle(file) as h5file:
            name = list(h5file)
            if len(name) != 1:
                raise FileNotFoundError

            attributes = {}

            for key, value in h5file[name[0]].attrs.items():
                attributes[key] = value

        return attributes

    @classmethod
    def fetch_property_groups(
        cls, file: str | h5py.File, uid: uuid.UUID
    ) -> dict[str, dict[str, str]]:
        r"""
        Get the property groups.

        :param file: :obj:`h5py.File` or name of the target geoh5 file
        :param uid: Unique identifier of the target entity

        :return property_group_attributes: :obj:`dict` of property groups
            and respective attributes.

        .. code-block:: python

            property_group = {
                "group_1": {"attribute": value, ...},
                ...,
                "group_N": {"attribute": value, ...},
            }
        """
        with fetch_h5_handle(file) as h5file:
            name = list(h5file)[0]
            property_groups: dict[str, dict[str, str]] = {}
            try:
                pg_handle = h5file[name]["Objects"][as_str_if_uuid(uid)][
                    "PropertyGroups"
                ]
                for pg_uid in pg_handle:
                    property_groups[pg_uid] = {}
                    for attr, value in pg_handle[pg_uid].attrs.items():
                        property_groups[pg_uid][INV_KEY_MAP.get(attr, attr)] = value
            except KeyError:
                pass
        return property_groups

    @classmethod
    def fetch_type(
        cls, file: str | h5py.File, uid: uuid.UUID, entity_type: str
    ) -> dict:
        """
        Fetch a type from the target geoh5.

        :param file: :obj:`h5py.File` or name of the target geoh5 file
        :param uid: Unique identifier of the target entity
        :param entity_type: One of 'Data', 'Object' or 'Group'
        :return property_group_attributes: :obj:`dict` of property groups
            and respective attributes.

        """
        with fetch_h5_handle(file) as h5file:
            name = list(h5file)[0]
            entity_type = entity_type + " types"
            type_handle = h5file[name]["Types"][entity_type][as_str_if_uuid(uid)]
            return cls.fetch_type_attributes(type_handle)

    @classmethod
    def fetch_type_attributes(cls, type_handle: h5py.Group) -> dict:
        """
        Fetch type attributes from a given h5 handle.
        """
        type_attributes = {}
        for key, value in type_handle.attrs.items():
            type_attributes[INV_KEY_MAP.get(key, key)] = value

        if "Color map" in type_handle:
            type_attributes["color_map"] = {}
            for key, value in type_handle["Color map"].attrs.items():
                type_attributes["color_map"][key] = value
            type_attributes["color_map"]["values"] = type_handle["Color map"][:]

        if "Value map" in type_handle:
            type_attributes["value_map"] = type_handle["Value map"][:].astype(
                [("Key", "<u4"), ("Value", h5py.special_dtype(vlen=str))]
            )
        return type_attributes

    @classmethod
    def fetch_uuids(cls, file: str | h5py.File, entity_type: str) -> list:
        """
        Fetch all uuids of a given type from geoh5

        :param file: :obj:`h5py.File` or name of the target geoh5 file
        :param entity_type: Type of entity from
            'group', 'data', 'object', 'group_type', 'data_type', 'object_type'

        :return uuids: [uuid1, uuid2, ...]
            List of uuids
        """
        with fetch_h5_handle(file) as h5file:
            name = list(h5file)[0]
            entity_type = cls.format_type_string(entity_type)
            try:
                uuids = [str2uuid(uid) for uid in h5file[name][entity_type]]
            except KeyError:
                uuids = []

        return uuids

    @classmethod
    def fetch_file_object(
        cls, file: str | h5py.File, uid: uuid.UUID, file_name: str
    ) -> bytes | None:
        """
        Load data associated with an image file

        :param file: Name of the target geoh5 file
        :param uid: Unique identifier of the target entity
        :param file_name: Name of the file stored as bytes data.

        :return values: Data file stored as bytes
        """
        with fetch_h5_handle(file) as h5file:
            name = list(h5file)[0]

            try:
                bytes_value = h5file[name]["Data"][as_str_if_uuid(uid)][file_name][
                    ()
                ].tobytes()

            except KeyError:
                bytes_value = None

        return bytes_value

    @classmethod
    def fetch_values(
        cls, file: str | h5py.File, uid: uuid.UUID
    ) -> np.ndarray | str | float | None:
        """
        Get data :obj:`~geoh5py.data.data.Data.values`

        :param file: :obj:`h5py.File` or name of the target geoh5 file
        :param uid: Unique identifier of the target entity

        :return values: :obj:`numpy.array` of :obj:`float`
        """
        with fetch_h5_handle(file) as h5file:
            name = list(h5file)[0]

            try:
                values = np.r_[h5file[name]["Data"][as_str_if_uuid(uid)]["Data"]]
                if isinstance(values[0], (str, bytes)):
                    values = np.asarray([as_str_if_utf8_bytes(val) for val in values])
                    if len(values) == 1:
                        values = values[0]
                else:
                    if values.dtype in [float, "float64", "float32"]:
                        ind = values == FLOAT_NDV
                        values[ind] = np.nan

            except KeyError:
                values = None

        return values

    @staticmethod
    def format_type_string(string: str) -> str:
        """Format names used for types."""
        string = string.capitalize()
        if string in ["Group", "Object"]:
            string += "s"

        return string



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from collections.abc import Callable

from .uijson import (
    ErrorPool,
    UIJsonError,
    dependency_type_validation,
    mesh_type_validation,
    parent_validation,
)


VALIDATIONS_MAP = {
    "dependency": dependency_type_validation,
    "mesh_type": mesh_type_validation,
    "parent": parent_validation,
}


def get_validations(form_keys: list[str]) -> list[Callable]:
    """Returns a list of callable validations based on identifying form keys."""
    return [VALIDATIONS_MAP[k] for k in form_keys if k in VALIDATIONS_MAP]



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

import json
import uuid

import numpy as np

from ..shared.utils import str_json_to_dict, stringify
from .base import Group


class UIJsonGroup(Group):
    """Group for SimPEG inversions."""

    _TYPE_UID = uuid.UUID("{BB50AC61-A657-4926-9C82-067658E246A0}")
    _default_name = "UIJson"

    def __init__(
        self,
        options: dict | np.ndarray | bytes | None = None,
        **kwargs,
    ):
        self._options: dict

        super().__init__(**kwargs)
        self.options = self.format_input_options(options)

    @property
    def options(self) -> dict:
        """
        Metadata attached to the entity.
        """
        return self._options

    @options.setter
    def options(self, value: dict):
        if not isinstance(value, dict):
            raise ValueError(f"Input 'options' must be of type {dict}.")

        self._options = value

        if self.on_file:
            self.workspace.update_attribute(self, "options")

    def add_ui_json(self, name: str | None = None):
        """
        Add ui.json file to entity.

        :param name: Name of the file in the workspace.
        """
        if self.options is None:
            raise ValueError("UIJsonGroup must have options set.")

        json_str = json.dumps(stringify(self.options), indent=4)
        bytes_data = json_str.encode("utf-8")

        if name is None:
            name = self.name

        self.add_file(bytes_data, name=f"{name}.ui.json")

    @staticmethod
    def format_input_options(options: dict | np.ndarray | bytes | None) -> dict:
        """
        Format input options to a dictionary.

        :param options: Input options.
        :return: Formatted options.
        """
        if options is None:
            return {}

        if isinstance(options, np.ndarray):
            options = options[0]

        if isinstance(options, bytes):
            options = str_json_to_dict(options)

        if not isinstance(options, dict):
            raise ValueError(f"Input 'options' must be of type {dict}.")

        return options



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


# pylint: disable=unused-import
# flake8: noqa

from .base import Group
from .container import ContainerGroup
from .custom import CustomGroup
from .drillhole import DrillholeGroup, IntegratorDrillholeGroup
from .giftools import GiftoolsGroup
from .group_type import GroupType
from .integrator import (
    AirborneTheme,
    BlastingDataSet,
    BlastingTheme,
    EarthModelsTheme,
    GeochemistryMineralogyDataSet,
    GeochemistryMineralogyTheme,
    GeophysicsTheme,
    GroundTheme,
    IncidentDataSet,
    IncidentsTheme,
    IntegratorGroup,
    IntegratorNeighbourhoods,
    IntegratorMaps,
    IntegratorMapFile,
    IntegratorProject,
    MineProductionTheme,
    ObservationPointsTheme,
    QueryGroup,
    RockPropertiesTheme,
    SamplesTheme,
)
from .interpretation_section import InterpretationSection
from .notype import NoTypeGroup
from .property_group import PropertyGroup
from .root import RootGroup
from .simpeg import SimPEGGroup
from .survey import AirborneGeophysics
from .uijson import UIJsonGroup
